"""
Definitions of the scaling algorithm.
"""

from __future__ import annotations

import itertools
import json
import logging
import time

from libtbx import Auto

from dials.algorithms.scaling.observers import (
    ScalingHTMLContextManager,
    ScalingSummaryContextManager,
)
from dials.algorithms.scaling.scale_and_filter import AnalysisResults, log_cycle_results
from dials.algorithms.scaling.scaler_factory import MultiScalerFactory, create_scaler
from dials.algorithms.scaling.scaling_library import (
    create_datastructures_for_reference_file,
    create_scaling_model,
    determine_best_unit_cell,
    merging_stats_from_scaled_array,
    scaled_data_as_miller_array,
    set_image_ranges_in_scaling_models,
)
from dials.algorithms.scaling.scaling_utilities import (
    DialsMergingStatisticsError,
    log_memory_usage,
)
from dials.algorithms.statistics.cc_half_algorithm import (
    CCHalfFromDials as deltaccscript,
)
from dials.array_family import flex
from dials.command_line.compute_delta_cchalf import phil_scope as deltacc_phil_scope
from dials.command_line.cosym import cosym
from dials.command_line.cosym import phil_scope as cosym_phil_scope
from dials.util.exclude_images import (
    exclude_image_ranges_for_scaling,
    get_valid_image_ranges,
)
from dials.util.multi_dataset_handling import (
    assign_unique_identifiers,
    parse_multiple_datasets,
    select_datasets_on_ids,
    update_imageset_ids,
)

logger = logging.getLogger("dials")


def prepare_input(params, experiments, reflections):
    """Perform checks on the data and prepare the data for scaling.

    Raises:
        ValueError - a range of checks are made, a ValueError may be raised
            for a number of reasons.

    """

    #### First exclude any datasets, before the dataset is split into
    #### individual reflection tables and expids set.
    if (
        params.dataset_selection.exclude_datasets
        or params.dataset_selection.use_datasets
    ):
        experiments, reflections = select_datasets_on_ids(
            experiments,
            reflections,
            params.dataset_selection.exclude_datasets,
            params.dataset_selection.use_datasets,
        )
        ids = flex.size_t()
        for r in reflections:
            ids.extend(r.experiment_identifiers().keys())
        logger.info(
            "\nDataset ids for retained datasets are: %s \n",
            ",".join(str(i) for i in ids),
        )

    #### Split the reflections tables into a list of reflection tables,
    #### with one table per experiment.
    logger.info(
        "Checking for the existence of a reflection table \n"
        "containing multiple datasets \n"
    )
    reflections = parse_multiple_datasets(reflections)
    logger.info(
        "Found %s reflection tables & %s experiments in total.",
        len(reflections),
        len(experiments),
    )

    if len(experiments) != len(reflections):
        raise ValueError(
            "Mismatched number of experiments and reflection tables found."
        )

    #### Sort out deprecated options
    if params.scaling_options.target_model or params.scaling_options.target_mtz:
        if params.scaling_options.reference:
            raise ValueError(
                "Can't specify reference in addition to target_mtz/target_model"
            )
        if params.scaling_options.target_model and params.scaling_options.target_mtz:
            raise ValueError(
                "Can only specify one of target_mtz/target_model (both deprecated, use reference=)"
            )
        if params.scaling_options.target_model:
            logger.warning(
                "Warning: target_model option is deprecated and will be removed, please use reference="
            )
            params.scaling_options.reference = params.scaling_options.target_model
            params.scaling_options.target_model = None
        elif params.scaling_options.target_mtz:
            logger.warning(
                "Warning: target_mtz option is deprecated and will be removed, please use reference="
            )
            params.scaling_options.reference = params.scaling_options.target_mtz
            params.scaling_options.target_mtz = None

    #### Assign experiment identifiers.
    experiments, reflections = assign_unique_identifiers(experiments, reflections)
    ids = itertools.chain.from_iterable(
        r.experiment_identifiers().keys() for r in reflections
    )
    logger.info("\nDataset ids are: %s \n", ",".join(str(i) for i in ids))

    for r in reflections:
        r.unset_flags(flex.bool(len(r), True), r.flags.bad_for_scaling)
        r.unset_flags(flex.bool(r.size(), True), r.flags.scaled)

    reflections, experiments = exclude_image_ranges_for_scaling(
        reflections, experiments, params.exclude_images
    )

    #### Allow checking of consistent indexing, useful for
    #### targeted / incremental scaling.
    if params.scaling_options.check_consistent_indexing:
        logger.info("Running dials.cosym to check consistent indexing:\n")
        cosym_params = cosym_phil_scope.extract()
        cosym_params.nproc = params.scaling_options.nproc
        cosym_instance = cosym(experiments, reflections, cosym_params)
        cosym_instance.run()
        experiments = cosym_instance.experiments
        reflections = cosym_instance.reflections
        logger.info("Finished running dials.cosym, continuing with scaling.\n")

    #### Make sure all experiments in same space group
    sgs = [expt.crystal.get_space_group().type().number() for expt in experiments]
    if len(set(sgs)) > 1:
        raise ValueError(
            """The experiments have different space groups:
            space group numbers found: %s
            Please reanalyse the data so that space groups are consistent,
            (consider using dials.reindex, dials.symmetry or dials.cosym) or
            remove incompatible experiments (using the option exclude_datasets=)"""
            % ", ".join(map(str, set(sgs)))
        )
    logger.info(
        "Space group being used during scaling is %s",
        experiments[0].crystal.get_space_group().info(),
    )

    #### If doing targeted scaling, extract data and append an experiment
    #### and reflection table to the lists
    if params.scaling_options.reference:
        # Set a suitable d_min in the case when we might have a model file
        d_min_for_structure_model = 2.0
        if params.cut_data.d_min not in (None, Auto):
            d_min_for_structure_model = params.cut_data.d_min
        expt, reflection_table = create_datastructures_for_reference_file(
            experiments,
            params.scaling_options.reference,
            params.anomalous,
            d_min=d_min_for_structure_model,
        )
        experiments.append(expt)
        reflections.append(reflection_table)

    #### Perform any non-batch cutting of the datasets, including the target dataset
    best_unit_cell = params.reflection_selection.best_unit_cell
    if best_unit_cell is None:
        best_unit_cell = determine_best_unit_cell(experiments)
    for reflection in reflections:
        if params.cut_data.d_min or params.cut_data.d_max:
            d = best_unit_cell.d(reflection["miller_index"])
            if params.cut_data.d_min:
                sel = d < params.cut_data.d_min
                reflection.set_flags(sel, reflection.flags.user_excluded_in_scaling)
            if params.cut_data.d_max:
                sel = d > params.cut_data.d_max
                reflection.set_flags(sel, reflection.flags.user_excluded_in_scaling)
        if params.cut_data.partiality_cutoff and "partiality" in reflection:
            reflection.set_flags(
                reflection["partiality"] < params.cut_data.partiality_cutoff,
                reflection.flags.user_excluded_in_scaling,
            )
    return params, experiments, reflections


class ScalingAlgorithm:
    def __init__(self, params, experiments, reflections):
        self.scaler = None
        self.scaled_miller_array = None
        self.merging_statistics_result = None
        self.anom_merging_statistics_result = None
        self.filtering_results = None
        self.params, self.experiments, self.reflections = prepare_input(
            params, experiments, reflections
        )
        self._create_model_and_scaler()
        logger.debug("Initialised scaling script object")
        log_memory_usage()

    def _create_model_and_scaler(self):
        """Create the scaling models and scaler."""
        self.experiments = create_scaling_model(
            self.params, self.experiments, self.reflections
        )
        logger.info("\nScaling models have been initialised for all experiments.")
        logger.info("%s%s%s", "\n", "=" * 80, "\n")

        self.experiments = set_image_ranges_in_scaling_models(self.experiments)

        self.scaler = create_scaler(self.params, self.experiments, self.reflections)

    def run(self):
        """Run the scaling script."""
        with ScalingHTMLContextManager(self), ScalingSummaryContextManager(self):
            start_time = time.time()
            self.scale()
            self.remove_bad_data()
            if not self.experiments:
                raise ValueError("All data sets have been rejected as bad.")
            for table in self.reflections:
                bad = table.get_flags(table.flags.bad_for_scaling, all=False)
                table.unset_flags(flex.bool(table.size(), True), table.flags.scaled)
                table.set_flags(~bad, table.flags.scaled)
            self.scaled_miller_array = scaled_data_as_miller_array(
                self.reflections,
                self.experiments,
                anomalous_flag=False,
                best_unit_cell=self.params.reflection_selection.best_unit_cell,
            )
            try:
                self.calculate_merging_stats()
            except DialsMergingStatisticsError as e:
                logger.info(e)

            # All done!
            logger.info("\nTotal time taken: %.4fs ", time.time() - start_time)
            logger.info("%s%s%s", "\n", "=" * 80, "\n")

    def scale(self):
        """The main scaling algorithm."""

        if self.scaler.id_ == "target":
            ### FIXME add in quick prescaling round if large scale difference?
            self.scaler.perform_scaling()

            if (
                self.params.scaling_options.only_target
                or self.params.scaling_options.reference
            ):

                self.scaler = targeted_scaling_algorithm(self.scaler)
                return
            # Now pass to a multiscaler ready for next round of scaling.
            self.scaler.expand_scales_to_all_reflections()
            self.scaler = MultiScalerFactory.create_from_targetscaler(self.scaler)

        # From here onwards, scaler should only be a SingleScaler
        # or MultiScaler (not TargetScaler).
        self.scaler = scaling_algorithm(self.scaler)

    def remove_bad_data(self):
        """Remove any target model/mtz data and any datasets which were removed
        from the scaler during scaling."""
        # first remove target refl/exps
        if (
            self.params.scaling_options.reference
            or self.params.scaling_options.only_target
        ):
            # now remove things that were used as the target:
            n_target = len(self.experiments) - len(self.scaler.active_scalers)
            self.experiments = self.experiments[:-n_target]
            self.reflections = self.reflections[:-n_target]
        # remove any bad datasets:
        removed_ids = self.scaler.removed_datasets
        if removed_ids:
            logger.info("deleting removed datasets from memory: %s", removed_ids)
            expids = list(self.experiments.identifiers())
            locs_in_list = [expids.index(expid) for expid in removed_ids]
            self.experiments, self.reflections = select_datasets_on_ids(
                self.experiments, self.reflections, exclude_datasets=locs_in_list
            )
        # also remove negative scales (or scales below 0.001)
        n = 0
        for table in self.reflections:
            bad_sf = (
                table["inverse_scale_factor"] < self.params.cut_data.small_scale_cutoff
            )
            n += bad_sf.count(True)
            table.set_flags(bad_sf, table.flags.excluded_for_scaling)
        if n > 0:
            logger.info(
                f"{n} reflections excluded: scale factor < {self.params.cut_data.small_scale_cutoff}"
            )

    def calculate_merging_stats(self):
        try:
            (
                self.merging_statistics_result,
                self.anom_merging_statistics_result,
            ) = merging_stats_from_scaled_array(
                self.scaled_miller_array,
                self.params.output.merging.nbins,
                self.params.output.use_internal_variance,
            )
        except DialsMergingStatisticsError as e:
            logger.warning(e, exc_info=True)

    def finish(self):
        """Save the experiments json and scaled pickle file."""

        # Now create a joint reflection table. Delete all other data before
        # joining reflection tables - just need experiments for mtz export
        # and a reflection table.
        del self.scaler
        cols_to_del = [
            "variance",
            "intensity",
            "s0",
            "s0c",
            "s1c",
            "prescaling_correction",
            "batch",
        ]
        for table in self.reflections:
            for col in cols_to_del:
                try:
                    del table[col]
                except KeyError:
                    pass

        # update imageset ids before combining reflection tables.
        self.reflections = update_imageset_ids(self.experiments, self.reflections)
        # Note, we don't use flex.reflection_table.concat below on purpose, so
        # that the dataset ids in the table are consistent from input to output
        # when datasets are removed, e.g. by filtering, exclude_datasets= etc.
        joint_table = flex.reflection_table()
        for i in range(len(self.reflections)):
            joint_table.extend(self.reflections[i])
            self.reflections[i] = 0  # del reference from initial list

        # remove reflections with very low scale factors
        sel = (
            joint_table["inverse_scale_factor"]
            < self.params.cut_data.small_scale_cutoff
        )
        good_sel = ~joint_table.get_flags(joint_table.flags.bad_for_scaling, all=False)
        n_low = (good_sel & sel).count(True)
        if n_low > 0:
            logger.warning(
                f"""{n_low} non-excluded reflections were assigned scale factors < {self.params.cut_data.small_scale_cutoff} during scaling.
These will be excluded in the output reflection table. It may be best to rerun
scaling from this point for an improved model."""
            )
            joint_table.set_flags(sel, joint_table.flags.excluded_for_scaling)

        return self.experiments, joint_table


class ScaleAndFilterAlgorithm(ScalingAlgorithm):
    def __init__(self, params, experiments, reflections):
        super().__init__(params, experiments, reflections)
        if (
            params.filtering.deltacchalf.mode == "dataset"
            and self.scaler.id_ != "multi"
        ):
            raise ValueError(
                """\
Whole dataset deltacchalf scaling and filtering can only be performed in
multi-dataset scaling mode (not single dataset or scaling against a reference)"""
            )

    def run(self):
        """Run cycles of scaling and filtering."""
        with ScalingHTMLContextManager(self):
            start_time = time.time()
            results = AnalysisResults()

            for counter in range(1, self.params.filtering.deltacchalf.max_cycles + 1):
                self.run_scaling_cycle()

                if counter == 1:
                    results.initial_expids_and_image_ranges = [
                        (exp.identifier, exp.scan.get_image_range())
                        if exp.scan
                        else None
                        for exp in self.experiments
                    ]

                delta_cc_params = deltacc_phil_scope.extract()
                delta_cc_params.mode = self.params.filtering.deltacchalf.mode
                delta_cc_params.group_size = (
                    self.params.filtering.deltacchalf.group_size
                )
                delta_cc_params.stdcutoff = self.params.filtering.deltacchalf.stdcutoff
                logger.info("\nPerforming a round of filtering.\n")

                # need to reduce to single table.
                joined_reflections = flex.reflection_table()
                for table in self.reflections:
                    joined_reflections.extend(table)

                script = deltaccscript(
                    delta_cc_params, self.experiments, joined_reflections
                )
                script.run()

                valid_image_ranges = get_valid_image_ranges(self.experiments)
                results.expids_and_image_ranges = [
                    (exp.identifier, valid_image_ranges[i]) if exp.scan else None
                    for i, exp in enumerate(self.experiments)
                ]

                self.experiments = script.experiments
                self.params.dataset_selection.use_datasets = None
                self.params.dataset_selection.exclude_datasets = None

                results = log_cycle_results(results, self, script)
                logger.info(
                    "Cycle %s of filtering, n_reflections removed this cycle: %s",
                    counter,
                    results.get_last_cycle_results()["n_removed"],
                )

                # Test termination conditions
                latest_results = results.get_last_cycle_results()
                if latest_results["n_removed"] == 0:
                    logger.info(
                        "Finishing scaling and filtering as no data removed in this cycle."
                    )
                    self.reflections = parse_multiple_datasets(
                        [script.filtered_reflection_table]
                    )
                    if self.params.scaling_options.full_matrix:
                        results = self._run_final_scale_cycle(results)
                    results.finish(termination_reason="no_more_removed")
                    break

                # Need to split reflections for further processing.
                self.reflections = parse_multiple_datasets(
                    [script.filtered_reflection_table]
                )

                if (
                    latest_results["cumul_percent_removed"]
                    > self.params.filtering.deltacchalf.max_percent_removed
                ):
                    logger.info(
                        "Finishing scale and filtering as have now removed more than the limit."
                    )
                    results = self._run_final_scale_cycle(results)
                    results.finish(termination_reason="max_percent_removed")
                    break

                if self.params.filtering.deltacchalf.min_completeness:
                    if (
                        latest_results["merging_stats"]["completeness"]
                        < self.params.filtering.deltacchalf.min_completeness
                    ):
                        logger.info(
                            "Finishing scaling and filtering as completeness now below cutoff."
                        )
                        results = self._run_final_scale_cycle(results)
                        results.finish(termination_reason="below_completeness_limit")
                        break

                if counter == self.params.filtering.deltacchalf.max_cycles:
                    logger.info("Finishing as reached max number of cycles.")
                    results = self._run_final_scale_cycle(results)
                    results.finish(termination_reason="max_cycles")
                    break

                # If not finished then need to create new scaler to try again
                self._create_model_and_scaler()
            self.filtering_results = results
            # Print summary of results
            logger.info(results)
            with open(self.params.filtering.output.scale_and_filter_results, "w") as f:
                json.dump(self.filtering_results.to_dict(), f, indent=2)
            # All done!
            logger.info("\nTotal time taken: %.4fs ", time.time() - start_time)
            logger.info("%s%s%s", "\n", "=" * 80, "\n")

    def run_scaling_cycle(self):
        """Do a round of scaling for scaling and filtering."""
        # Turn off the full matrix round, all else is the same.

        initial_full_matrix = self.params.scaling_options.full_matrix
        self.scaler.params.scaling_options.full_matrix = False
        self.scaler = scaling_algorithm(self.scaler)
        self.scaler.params.scaling_options.full_matrix = initial_full_matrix
        self.remove_bad_data()
        for table in self.reflections:
            bad = table.get_flags(table.flags.bad_for_scaling, all=False)
            table.unset_flags(flex.bool(table.size(), True), table.flags.scaled)
            table.set_flags(~bad, table.flags.scaled)
        self.scaled_miller_array = scaled_data_as_miller_array(
            self.reflections,
            self.experiments,
            anomalous_flag=False,
            best_unit_cell=self.params.reflection_selection.best_unit_cell,
        )
        try:
            self.calculate_merging_stats()
        except DialsMergingStatisticsError as e:
            logger.info(e)
        logger.info("Performed cycle of scaling.")

    def _run_final_scale_cycle(self, results):
        self._create_model_and_scaler()
        super().run()
        results.add_final_stats(self.merging_statistics_result)
        for table in self.reflections:
            bad = table.get_flags(table.flags.bad_for_scaling, all=False)
            table.unset_flags(flex.bool(table.size(), True), table.flags.scaled)
            table.set_flags(~bad, table.flags.scaled)
        return results


def expand_and_do_outlier_rejection(scaler, calc_cov=False):
    """Calculate scales for all reflections and do outlier rejection."""
    scaler.expand_scales_to_all_reflections(calc_cov=calc_cov)
    if scaler.params.scaling_options.outlier_rejection:
        scaler.round_of_outlier_rejection()


def do_intensity_combination(scaler, reselect=True):
    """
    Do prf/sum intensity combination.

    Optionally reselect reflections to prepare for another minimisation round.
    """
    if scaler.params.reflection_selection.intensity_choice == "combine":
        scaler.combine_intensities()
        if scaler.params.scaling_options.outlier_rejection:
            scaler.round_of_outlier_rejection()
    if reselect:
        scaler.make_ready_for_scaling()


def do_error_analysis(scaler, reselect=True):
    """
    Do error model analysis.

    Optionally reselect reflections to prepare for another minimisation round.
    """
    if scaler.params.weighting.error_model.error_model:
        scaler.perform_error_optimisation()
    if reselect:
        scaler.make_ready_for_scaling()


def scaling_algorithm(scaler):
    """Main algorithm for scaling."""
    scaler.perform_scaling()
    need_to_rescale = False

    if (
        scaler.params.reflection_selection.intensity_choice == "combine"
        or scaler.params.scaling_options.outlier_rejection
    ):

        expand_and_do_outlier_rejection(scaler)

        do_intensity_combination(scaler, reselect=True)

        need_to_rescale = True

    if (
        scaler.params.weighting.error_model.error_model
        or scaler.params.scaling_options.outlier_rejection
    ):
        if need_to_rescale:
            scaler.perform_scaling()

        expand_and_do_outlier_rejection(scaler)

        do_error_analysis(scaler, reselect=True)

        need_to_rescale = True

    if scaler.params.scaling_options.full_matrix:

        scaler.perform_scaling(
            engine=scaler.params.scaling_refinery.full_matrix_engine,
            max_iterations=scaler.params.scaling_refinery.full_matrix_max_iterations,
        )
        # check if we're fixing a parameter, if so, redo full matrix with
        # smaller tolerance for one cycle.
        need_to_scale = scaler.fix_initial_parameter()
        if need_to_scale:
            scaler.perform_scaling(
                engine=scaler.params.scaling_refinery.full_matrix_engine,
                max_iterations=1,
                tolerance=scaler.params.scaling_refinery.rmsd_tolerance / 4.0,
            )
    elif need_to_rescale:
        scaler.perform_scaling()

    # The minimisation has only been done on a subset on the data, so apply the
    # scale factors to the whole reflection table.

    scaler.clear_Ih_table()
    expand_and_do_outlier_rejection(scaler, calc_cov=True)
    do_error_analysis(scaler, reselect=False)

    scaler.prepare_reflection_tables_for_output()
    return scaler


def targeted_scaling_algorithm(scaler):
    """Main algorithm for targeted scaling."""

    if scaler.params.scaling_options.outlier_rejection:
        expand_and_do_outlier_rejection(scaler)
        scaler.make_ready_for_scaling()
        scaler.perform_scaling()

    if scaler.params.scaling_options.full_matrix and (
        scaler.params.scaling_refinery.engine == "SimpleLBFGS"
    ):
        scaler.perform_scaling(
            engine=scaler.params.scaling_refinery.full_matrix_engine,
            max_iterations=scaler.params.scaling_refinery.full_matrix_max_iterations,
        )

    expand_and_do_outlier_rejection(scaler, calc_cov=True)
    # do_error_analysis(scaler, reselect=False)

    scaler.prepare_reflection_tables_for_output()
    return scaler



"""
Tests for the constraints system used in refinement
"""

from __future__ import annotations

from copy import deepcopy
from pathlib import Path

import procrunner

from dxtbx.model.experiment_list import ExperimentListFactory
from libtbx.test_utils import approx_equal
from scitbx import sparse

from dials.algorithms.refinement.constraints import (
    ConstraintManager,
    EqualShiftConstraint,
    SparseConstraintManager,
)
from dials.algorithms.refinement.engine import Journal
from dials.array_family import flex


def test_contraints_manager_simple_test():

    x = flex.random_double(10)

    # constrain parameters 2 and 4 and 6, 7 and 8
    c1 = EqualShiftConstraint([1, 3], x)
    c2 = EqualShiftConstraint([5, 6, 7], x)

    cm = ConstraintManager([c1, c2], len(x))
    constrained_x = cm.constrain_parameters(x)

    # check the constrained parameters are as expected
    assert len(constrained_x) == 7
    assert constrained_x[5] == flex.mean(x.select([1, 3]))
    assert constrained_x[6] == flex.mean(x[5:8])

    # minimiser would modify the constrained parameters
    mod_constrained_x = constrained_x + 10.0

    # check the expanded parameters are as expected
    expanded = cm.expand_parameters(mod_constrained_x)
    assert x + 10.0 == expanded

    # make a matrix to exercise jacobian compaction
    j = flex.random_double(20 * 10)
    j.reshape(flex.grid(20, 10))

    # for constrained columns, elements that are non-zero in one column are
    # zero in the other columns. Enforce that in this example
    mask2 = flex.bool([True] * 10 + [False] * 10)
    mask4 = ~mask2
    col2 = j.matrix_copy_column(1)
    col2.set_selected(mask2, 0)
    j.matrix_paste_column_in_place(col2, 1)
    col4 = j.matrix_copy_column(3)
    col4.set_selected(mask4, 0)
    j.matrix_paste_column_in_place(col4, 3)

    mask6 = flex.bool([False] * 7 + [True] * 13)
    mask7 = mask6.reversed()
    mask8 = ~(mask6 & mask7)
    col6 = j.matrix_copy_column(5)
    col6.set_selected(mask6, 0)
    j.matrix_paste_column_in_place(col6, 5)
    col7 = j.matrix_copy_column(6)
    col7.set_selected(mask7, 0)
    j.matrix_paste_column_in_place(col7, 6)
    col8 = j.matrix_copy_column(7)
    col8.set_selected(mask8, 0)
    j.matrix_paste_column_in_place(col8, 7)

    cj = cm.constrain_jacobian(j)

    # check expected dimensions
    assert cj.all() == (20, 7)

    # check that the constrained columns are equal to sums of the relevant
    # columns in the original Jacobian
    tmp = j.matrix_copy_column(1) + j.matrix_copy_column(3)
    assert (cj.matrix_copy_column(5) == tmp).all_eq(True)

    tmp = j.matrix_copy_column(5) + j.matrix_copy_column(6) + j.matrix_copy_column(7)
    assert (cj.matrix_copy_column(6) == tmp).all_eq(True)

    # convert to a sparse matrix to exercise the sparse Jacobian compaction
    j2 = sparse.matrix(20, 10)
    mask = flex.bool(20, True)
    for i, c in enumerate(j2.cols()):
        c.set_selected(mask, j.matrix_copy_column(i))
    assert (j2.as_dense_matrix() == j).all_eq(True)

    cm2 = SparseConstraintManager([c1, c2], len(x))
    cj2 = cm2.constrain_jacobian(j2)

    # ensure dense and sparse calculations give the same result
    assert (cj2.as_dense_matrix() == cj).all_eq(True)

    # construct derivatives of the objective dL/dp from the Jacobian to test
    # constrain_gradient_vector. Here assume unit weights
    dL_dp = [sum(col.as_dense_vector()) for col in j2.cols()]
    constr_dL_dp = cm.constrain_gradient_vector(dL_dp)

    # check constrained values are equal to sums of relevant elements in the
    # original gradient vector
    assert constr_dL_dp[5] == dL_dp[1] + dL_dp[3]
    assert constr_dL_dp[6] == dL_dp[5] + dL_dp[6] + dL_dp[7]


def test_constrained_refinement(dials_regression, tmp_path):
    """Test joint refinement where two detectors are constrained to enforce a
    differential distance (along the shared initial normal vector) of 1 mm.
    This test can be constructed on the fly from data already in
    dials_regression"""

    # use the 'centroid' data for this test. The 'regularized' experiments are
    # useful because the detector has fast and slow exactly aligned with X, -Y
    # so the distance is exactly along the normal vector and can be altered
    # directly by changing the Z component of the origin vector
    data_dir = Path(dials_regression) / "refinement_test_data" / "centroid"
    experiments_path = data_dir / "experiments_XPARM_REGULARIZED.json"
    pickle_path = data_dir / "spot_1000_xds.pickle"

    # load the experiments and spots
    el = ExperimentListFactory.from_json_file(experiments_path, check_format=False)
    rt = flex.reflection_table.from_file(pickle_path)

    # adjust the detector distance by -0.5 mm
    detector = el[0].detector
    panel = detector[0]
    fast = panel.get_fast_axis()
    slow = panel.get_slow_axis()
    origin = panel.get_origin()
    panel.set_frame(fast, slow, origin[0:2] + (origin[2] + 0.5,))

    # duplicate the experiment and adjust distance by +1 mm
    e2 = deepcopy(el[0])
    detector = e2.detector
    panel = detector[0]
    fast = panel.get_fast_axis()
    slow = panel.get_slow_axis()
    origin = panel.get_origin()
    panel.set_frame(fast, slow, origin[0:2] + (origin[2] - 1.0,))

    # append to the experiment list and write out
    el.append(e2)
    el.as_json(tmp_path / "foo.expt")

    # duplicate the reflections and increment the experiment id
    rt2 = deepcopy(rt)
    rt2["id"] = rt2["id"] + 1

    # concatenate reflections and write out
    rt.extend(rt2)
    rt.as_file(tmp_path / "foo.refl")

    # set up refinement, constraining the distance parameter
    result = procrunner.run(
        (
            "dials.refine",
            "foo.expt",
            "foo.refl",
            "history=history.json",
            "refinement.parameterisation.detector.constraints.parameter=Dist",
            "scan_varying=False",
        ),
        working_directory=tmp_path,
    )
    assert result.returncode == 0 and not result.stderr

    # load refinement history
    history = Journal.from_json_file(tmp_path / "history.json")
    ref_exp = ExperimentListFactory.from_json_file(
        tmp_path / "refined.expt", check_format=False
    )

    # we expect 8 steps of constrained refinement
    assert history.get_nrows() == 8

    # get parameter vector from the final step
    pvec = history["parameter_vector"][-1]

    # the constrained parameters have indices 0 and 6 in this case. Check they
    # are still exactly 1 mm apart
    assert pvec[0] == pvec[6] - 1.0

    # NB because the other detector parameters were not also constrained, the
    # refined lab frame distances may not in fact differ by 1 mm. The constraint
    # acts along the initial detector normal vector during composition of a new
    # detector position. After refinement of tilt/twist type rotations,
    # the final distances along the new normal vectors will change
    det1, det2 = ref_exp.detectors()
    p1 = det1[0]
    p2 = det2[0]
    assert approx_equal(p2.get_distance() - p1.get_distance(), 0.9987655)



from __future__ import annotations


class GLMBackgroundExt:
    """An extension class implementing a robust GLM background algorithm."""

    name = "glm"

    @staticmethod
    def phil():
        from libtbx.phil import parse

        phil = parse(
            """

      robust {
        tuning_constant = 1.345
          .type = float
          .help = "The tuning constant for robust estimation"
      }

      model {
        algorithm = constant2d *constant3d loglinear2d loglinear3d
          .type = choice
          .help = "The background model to fit"
      }

      min_pixels = 10
        .type = int(value_min=1)
        .help = "The minimum number of pixels required"
    """
        )
        return phil

    def __init__(self, params, experiments):
        """
        Initialise the algorithm.

        :param params: The input parameters
        :param experiments: The list of experiments
        """
        from libtbx.phil import parse

        from dials.algorithms.background.glm import BackgroundAlgorithm

        # Create some default parameters
        if params is None:
            params = self.phil().fetch(parse("")).extract()
        else:
            params = params.integration.background.glm

        # Create the algorithm
        self._algorithm = BackgroundAlgorithm(
            experiments,
            tuning_constant=params.robust.tuning_constant,
            model=params.model.algorithm,
            min_pixels=params.min_pixels,
        )

    def compute_background(self, reflections, image_volume=None):
        """
        Compute the background.

        :param reflections: The list of reflections
        """
        return self._algorithm.compute_background(
            reflections, image_volume=image_volume
        )



from __future__ import annotations

import os

import pytest

from dials.util.reference import (
    intensities_from_reference_data_file,
    intensities_from_reference_file,
    intensities_from_reference_model_file,
)


def test_intensities_from_reference_model(dials_data):
    "Test importing from a pdb/cif model"

    pdb_file = os.fspath(dials_data("cunir_serial", pathlib=True) / "2BW4.pdb")
    intensities = intensities_from_reference_file(pdb_file)
    assert intensities.data()
    assert not intensities.anomalous_flag()

    i2 = intensities_from_reference_model_file(pdb_file, wavelength=1.0)
    assert i2.data()
    assert i2.anomalous_flag()

    bad_input = "bad.pdbb"
    with pytest.raises(ValueError):
        _ = intensities_from_reference_file(bad_input)
    with pytest.raises(ValueError):
        _ = intensities_from_reference_model_file(bad_input)

    # Now try with the cif model
    cif_file = os.fspath(dials_data("cunir_serial", pathlib=True) / "2bw4.cif")
    i3 = intensities_from_reference_file(cif_file, wavelength=1.0)
    assert i3.data()
    assert i3.anomalous_flag()


def test_intensities_from_reference_data_file(dials_data):
    "Test importing from an mtz/cif datafile"

    mtz_file = os.fspath(
        dials_data("x4wide_processed", pathlib=True) / "AUTOMATIC_DEFAULT_scaled.mtz"
    )
    intensities = intensities_from_reference_file(mtz_file)
    assert intensities.data()
    assert intensities.anomalous_flag()

    i2 = intensities_from_reference_data_file(mtz_file)
    assert i2.data() == intensities.data()

    bad_input = "bad.mtzz"
    with pytest.raises(ValueError):
        _ = intensities_from_reference_data_file(bad_input)

    # Now try with the cif data file from PDB entry 2BW4
    cif_file = os.fspath(dials_data("cunir_serial", pathlib=True) / "2bw4-sf.cif")
    i3 = intensities_from_reference_file(cif_file)
    assert i3.data()
    # In this file, the data is not anomalous
    assert not i3.anomalous_flag()



from __future__ import annotations

import os
import pickle

import pytest

from dxtbx.format.Format import Reader
from dxtbx.imageset import ImageSet, ImageSetData
from dxtbx.model.beam import Beam
from dxtbx.model.detector import Detector
from dxtbx.model.experiment_list import ExperimentListFactory
from libtbx import easy_run
from scitbx import matrix


def make_detector():
    """Make a dummy 4 panel detector with not many pixels to ensure test runs
    quickly"""
    pixel_size_x = 0.1
    pixel_size_y = 0.1
    npixels_per_panel_x = 50
    npixels_per_panel_y = 50
    distance = 100
    fast = matrix.col((1, 0, 0))
    slow = matrix.col((0, -1, 0))
    s0u = matrix.col((0, 0, -1))

    shift_x = -1.0 * npixels_per_panel_x * pixel_size_x * fast
    shift_y = -1.0 * npixels_per_panel_y * pixel_size_y * slow
    beam_centre = distance * s0u
    orig = beam_centre + shift_x + shift_y

    d = Detector()
    root = d.hierarchy()
    root.set_local_frame(fast.elems, slow.elems, orig.elems)

    local_fast = matrix.col((1.0, 0.0, 0.0))
    local_slow = matrix.col((0.0, 1.0, 0.0))
    span_x = npixels_per_panel_x * pixel_size_x * local_fast
    span_y = npixels_per_panel_y * pixel_size_y * local_slow

    local_origins = [matrix.col((0, 0, 0)), span_x, span_y, span_x + span_y]

    for i, lo in enumerate(local_origins):
        p = d.add_panel()
        p.set_image_size((npixels_per_panel_x, npixels_per_panel_y))
        p.set_pixel_size((pixel_size_x, pixel_size_y))
        p.set_local_frame(local_fast.elems, local_slow.elems, lo.elems)

    return d


def test_translate(dials_regression, run_in_tmp_path):
    """Test as written in https://github.com/dials/dials/issues/471. This
    is pretty slow!"""

    # use the i04_weak_data for this test
    data_dir = os.path.join(dials_regression, "image_examples", "DLS_I04")
    image_path = os.path.join(data_dir, "grid_full_cbf_0005.cbf")

    # Generate distortion maps
    cmd = f"dials.generate_distortion_maps {image_path} dx=1 dy=2"
    easy_run.fully_buffered(command=cmd).raise_if_errors()

    # Import without correction
    cmd = f"dials.import {image_path}"
    easy_run.fully_buffered(command=cmd).raise_if_errors()
    expt1 = ExperimentListFactory.from_serialized_format("imported.expt")[0]

    # Should be no dx/dy lookup files
    assert not expt1.imageset.external_lookup.dx.filename
    assert not expt1.imageset.external_lookup.dy.filename

    # Import with correction
    cmd = f"dials.import {image_path} dx=dx.pickle dy=dy.pickle output.experiments=corrected.expt"

    easy_run.fully_buffered(command=cmd).raise_if_errors()
    expt2 = ExperimentListFactory.from_serialized_format("corrected.expt")[0]

    # Check that dx/dy lookup files have been set
    assert expt2.imageset.external_lookup.dx.filename
    assert expt2.imageset.external_lookup.dy.filename

    # FIXME finish test by comparing px to mm positions for expt1.detector
    # and expt2.detector


def test_elliptical_distortion(run_in_tmp_path):
    """Create distortion maps for elliptical distortion using a dummy experiments
    with a small detector, for speed. Check those maps seem sensible"""

    # Make a detector model
    d = make_detector()

    # The beam is also essential for a experiments to be serialisable
    b = Beam((0, 0, 1), 1.0)

    # Create and write out a experiments
    imageset = ImageSet(ImageSetData(Reader(None, ["non-existent.cbf"]), None))
    imageset.set_detector(d)
    imageset.set_beam(b)
    experiments = ExperimentListFactory.from_imageset_and_crystal(imageset, None)
    experiments.as_json("dummy.expt")

    # Centre of distortion will be the far corner from the origin of the first
    # panel
    centre_xy = d[0].get_image_size_mm()

    # Generate distortion maps
    cmd = (
        "dials.generate_distortion_maps dummy.expt "
        "mode=ellipse centre_xy={},{} "
        "phi=0 l1=1.0 l2=0.95"
    ).format(*centre_xy)
    easy_run.fully_buffered(command=cmd).raise_if_errors()

    # Load the maps
    with open("dx.pickle", "rb") as f:
        dx = pickle.load(f)
    with open("dy.pickle", "rb") as f:
        dy = pickle.load(f)

    # Check there are 4 maps each
    assert len(dx) == len(dy) == 4

    # Ellipse has phi=0, so all correction is in the dy map
    for arr in dx:
        assert min(arr) == max(arr) == 0.0

    # The ellipse correction is centred at the middle of the detector and all in
    # the Y direction. Therefore we expect a few things from the dy maps:
    #
    # (1) Within each panel the columns of the array are identical.
    # (2) The two upper panels should be the same
    # (3) The two lower panels should be the same.
    # (4) One column from an upper panel is a negated, reversed column from a
    #     lower panel.
    #
    # All together expect the 4 dy maps to look something like this:
    #
    # /-----------\ /-----------\
    # |-3 -3 -3 -3| |-3 -3 -3 -3|
    # |-2 -2 -2 -2| |-2 -2 -2 -2|
    # |-1 -1 -1 -1| |-1 -1 -1 -1|
    # | 0  0  0  0| | 0  0  0  0|
    # \-----------/ \-----------/
    # /-----------\ /-----------\
    # | 0  0  0  0| | 0  0  0  0|
    # | 1  1  1  1| | 1  1  1  1|
    # | 2  2  2  2| | 2  2  2  2|
    # | 3  3  3  3| | 3  3  3  3|
    # \-----------/ \-----------/

    # So the fundamental data is all in the first column of first panel's map
    col0 = dy[0].matrix_copy_column(0)

    # The correction should be 5% of the distance from the ellipse centre to a
    # corrected pixel (l2 = 0.95 above) along the slow axis. Check that is the
    # case (for the first pixel at least)
    vec_centre_to_first_px = matrix.col(
        d[0].get_pixel_lab_coord((0.5, 0.5))
    ) - matrix.col(d[0].get_lab_coord(centre_xy))
    dist_centre_to_first_px = vec_centre_to_first_px.dot(
        matrix.col(d[0].get_slow_axis())
    )
    corr_mm = dist_centre_to_first_px * 0.05
    corr_px = corr_mm / d[0].get_pixel_size()[1]
    assert col0[0] == pytest.approx(corr_px)

    # Test (1) from above list for panel 0
    for i in range(1, 50):
        assert (col0 == dy[0].matrix_copy_column(i)).all_eq(True)

    # Test (2)
    assert (dy[0] == dy[1]).all_eq(True)

    # Test (3)
    assert (dy[2] == dy[3]).all_eq(True)

    # Test (4)
    assert col0 == pytest.approx(-1.0 * dy[2].matrix_copy_column(0).reversed())

    # Test (1) for panel 2 as well, which then covers everything needed
    col0 = dy[2].matrix_copy_column(0)
    for i in range(1, 50):
        assert (col0 == dy[2].matrix_copy_column(i)).all_eq(True)



from __future__ import annotations

from pathlib import Path

import procrunner

from dxtbx.model.experiment_list import ExperimentListFactory
from scitbx import matrix


def test1(dials_regression, tmp_path):
    """
    Refinement test of 300 CSPAD images, testing auto_reduction, parameter
    fixing, constraints, SparseLevMar, and sauter_poon outlier rejection. See
    README in the regression folder for more details.
    """
    data_dir = Path(dials_regression) / "refinement_test_data" / "cspad_refinement"

    result = procrunner.run(
        [
            "dials.refine",
            data_dir / "cspad_refined_experiments_step6_level2_300.json",
            data_dir / "cspad_reflections_step7_300.pickle",
            data_dir / "refine.phil",
        ],
        working_directory=tmp_path,
    )
    assert not result.returncode and not result.stderr

    # load results
    reg_exp = ExperimentListFactory.from_json_file(
        data_dir / "regression_experiments.json", check_format=False
    )
    ref_exp = ExperimentListFactory.from_json_file(
        tmp_path / "refined.expt", check_format=False
    )

    # compare results
    tol = 1e-5
    for b1, b2 in zip(reg_exp.beams(), ref_exp.beams()):
        assert b1.is_similar_to(
            b2,
            wavelength_tolerance=tol,
            direction_tolerance=tol,
            polarization_normal_tolerance=tol,
            polarization_fraction_tolerance=tol,
        )
        s0_1 = matrix.col(b1.get_unit_s0())
        s0_2 = matrix.col(b2.get_unit_s0())
        assert s0_1.accute_angle(s0_2, deg=True) < 0.0057  # ~0.1 mrad
    for c1, c2 in zip(reg_exp.crystals(), ref_exp.crystals()):
        assert c1.is_similar_to(c2)

    for d1, d2 in zip(reg_exp.detectors(), ref_exp.detectors()):
        assert d1.is_similar_to(
            d2,
            fast_axis_tolerance=1e-4,
            slow_axis_tolerance=1e-4,
            origin_tolerance=1e-2,
        )



from __future__ import annotations

import wx

_scores = {}


class ScoreSettingsFrame(wx.MiniFrame):
    # XXX Is this scoring or rating?

    def __init__(self, *args, **kwds):
        super().__init__(*args, **kwds)
        szr = wx.BoxSizer(wx.VERTICAL)
        panel = ScoreSettingsPanel(self)
        self.SetSizer(szr)
        szr.Add(panel, 1, wx.EXPAND)
        szr.Fit(panel)
        self.panel = panel
        self.sizer = szr
        self.Fit()
        self.Bind(wx.EVT_CLOSE, lambda evt: self.Destroy(), self)


class ScoreSettingsPanel(wx.Panel):
    def __init__(self, *args, **kwds):
        from wxtbx import bitmaps

        super().__init__(*args, **kwds)

        # Needed for communication with the root frame.
        self._root_frame = self.GetParent().GetParent()

        sizer = wx.BoxSizer(wx.VERTICAL)
        self.SetSizer(sizer)

        self._id_text = wx.NewId()
        text = wx.StaticText(self, id=self._id_text, label="")
        sizer.Add(text, flag=wx.ALIGN_CENTER)

        # The score buttons.  XXX Would really like this to be stars or
        # some such.  Could then zap the static text above.
        box = wx.BoxSizer(wx.HORIZONTAL)
        self._id_buttons = []
        for i in range(6):
            btn = wx.Button(self, label="%d" % i)
            box.Add(btn, flag=wx.ALIGN_CENTER, border=5)
            self.Bind(wx.EVT_BUTTON, self.OnScore, btn)
            self._id_buttons.append(btn.GetId())
        sizer.Add(box)

        # XXX Would like to have label and short help for the buttons.
        # Buttons in wxWidgets version 2.9.1 and later support text and
        # bitmap, see SetBitmap, SetBitmapLabel, SetBitmapDisabled, etc.
        box = wx.BoxSizer(wx.HORIZONTAL)

        self._id_previous = wx.NewId()
        btn = wx.BitmapButton(
            self,
            id=self._id_previous,
            bitmap=bitmaps.fetch_icon_bitmap("actions", "1leftarrow"),
            style=wx.BORDER_NONE,
        )
        box.Add(btn, flag=wx.ALIGN_CENTER, border=5)
        self.Bind(wx.EVT_BUTTON, self.OnPrevious, btn)

        btn = wx.Button(self, label="Save scores")
        box.Add(btn, flag=wx.ALIGN_CENTER, border=5)
        self.Bind(wx.EVT_BUTTON, self.OnSave, btn)

        self._id_next = wx.NewId()
        btn = wx.BitmapButton(
            self,
            id=self._id_next,
            bitmap=bitmaps.fetch_icon_bitmap("actions", "1rightarrow"),
            style=wx.BORDER_NONE,
        )
        box.Add(btn, flag=wx.ALIGN_CENTER, border=5)
        self.Bind(wx.EVT_BUTTON, self.OnNext, btn)

        sizer.Add(box, flag=wx.ALIGN_CENTER)

        # Register update events for the dynamic widgets.
        self.Bind(wx.EVT_UPDATE_UI, self.OnUpdateNext, id=self._id_next)
        self.Bind(wx.EVT_UPDATE_UI, self.OnUpdatePrevious, id=self._id_previous)
        self.Bind(wx.EVT_UPDATE_UI, self.OnUpdateText, id=self._id_text)

        for i in range(self._root_frame.image_chooser.GetCount()):
            _scores[
                self._root_frame.get_key(
                    self._root_frame.image_chooser.GetClientData(i)
                )
            ] = None

        self.Bind(wx.EVT_CHAR_HOOK, self.OnChar)

    def OnChar(self, event: wx.KeyEvent) -> None:
        keycode: int = event.GetKeyCode()

        key = self._get_image_key()
        if 0 <= keycode - ord("0") < len(self._id_buttons):
            _scores[key] = keycode - ord("0")
            print(f"Scoring {key}={_scores[key]}")
            self._root_frame.OnNext(event)
        elif keycode == ord(" "):
            # Count spacebar as zero, as 0 is far away
            _scores[key] = 0
            print(f"Scoring {key}={_scores[key]}")
            self._root_frame.OnNext(event)
        elif keycode == wx.WXK_LEFT:
            self._root_frame.OnPrevious(event)
        elif keycode == wx.WXK_RIGHT:
            self._root_frame.OnNext(event)

    def OnNext(self, event):
        self._root_frame.OnNext(event)

    def OnPrevious(self, event):
        self._root_frame.OnPrevious(event)

    def _get_image_key(self) -> str:
        return self._root_frame.get_key(
            self._root_frame.image_chooser.GetClientData(
                self._root_frame.image_chooser.GetSelection()
            )
        )

    def OnScore(self, event):
        score = self._id_buttons.index(event.EventObject.GetId())
        _scores[self._get_image_key()] = score
        self.OnNext(event)

    def OnSave(self, event):
        dialog = wx.FileDialog(
            self,
            defaultDir="",
            message="Save scores",
            style=wx.FD_SAVE | wx.FD_OVERWRITE_PROMPT,
            wildcard="Text files (*.txt)|*.txt",
        )
        if dialog.ShowModal() == wx.ID_OK:
            path = dialog.GetPath()
            if path != "":
                stream = open(path, "w")
                for (key, score) in _scores.items():
                    if score is None:
                        print(f"{key} None", file=stream)
                    else:
                        print("%s %d" % (key, score), file=stream)
                stream.close()
                print("Dumped scores to", path)

    def OnUpdateNext(self, event):
        root_next = self._root_frame.toolbar.FindById(wx.ID_FORWARD)
        event.Enable(root_next.IsEnabled())

    def OnUpdatePrevious(self, event):
        root_previous = self._root_frame.toolbar.FindById(wx.ID_BACKWARD)
        event.Enable(root_previous.IsEnabled())

    def OnUpdateText(self, event):
        key = self._root_frame.GetTitle()
        if key in _scores:
            event.SetText("Previous score: %d" % _scores[key])
        else:
            event.SetText("Not previously scored")

        # Call self.Layout() to recenter the updated text.
        self.Layout()



from __future__ import annotations

import procrunner

from dials.array_family import flex


def test_sort_intensities(dials_data, tmp_path):
    result = procrunner.run(
        [
            "dials.sort_reflections",
            dials_data("centroid_test_data", pathlib=True) / "integrated.pickle",
            "key=intensity.sum.value",
            "output=sorted1.refl",
        ],
        working_directory=tmp_path,
    )
    assert not result.returncode and not result.stderr
    assert tmp_path.joinpath("sorted1.refl").is_file()

    data = flex.reflection_table.from_file(tmp_path / "sorted1.refl")
    assert_sorted(data["intensity.sum.value"])


def test_reverse_sort_intensities(dials_data, tmp_path):
    result = procrunner.run(
        [
            "dials.sort_reflections",
            dials_data("centroid_test_data", pathlib=True) / "integrated.pickle",
            "output=sorted2.refl",
            "key=intensity.sum.value",
            "reverse=True",
        ],
        working_directory=tmp_path,
    )
    assert not result.returncode and not result.stderr
    assert tmp_path.joinpath("sorted2.refl").is_file()

    data = flex.reflection_table.from_file(tmp_path / "sorted2.refl")
    assert_sorted(data["intensity.sum.value"], reverse=True)


def test_default_sort_on_miller_index(dials_data, tmp_path):
    result = procrunner.run(
        [
            "dials.sort_reflections",
            dials_data("centroid_test_data", pathlib=True) / "integrated.pickle",
            "output=sorted3.refl",
        ],
        working_directory=tmp_path,
    )
    assert not result.returncode and not result.stderr
    assert tmp_path.joinpath("sorted3.refl").is_file()

    data = flex.reflection_table.from_file(tmp_path / "sorted3.refl")
    mi1 = data["miller_index"]
    orig = flex.reflection_table.from_file(
        dials_data("centroid_test_data", pathlib=True) / "integrated.pickle"
    )
    mi2 = flex.miller_index(sorted(orig["miller_index"]))
    assert mi1.all_eq(mi2)


def test_default_sort_on_miller_index_verbose(dials_data, tmp_path):
    result = procrunner.run(
        [
            "dials.sort_reflections",
            dials_data("centroid_test_data", pathlib=True) / "integrated.pickle",
            "output=sorted4.refl",
            "-v",
        ],
        working_directory=tmp_path,
    )
    assert not result.returncode and not result.stderr
    assert tmp_path.joinpath("sorted4.refl").is_file()
    assert b"Head of sorted list miller_index:" in result.stdout


def assert_sorted(data, reverse=False):
    assert len(data) > 0
    elem = data[0]
    for x in data:
        if reverse is True:
            assert x <= elem
        else:
            assert x >= elem
        elem = x



from __future__ import annotations

from libtbx.phil import parse

phil_scope = parse(
    """

lookup
  .help = "Parameters specifying lookup file path"
{
  mask = None
    .help = "The path to the mask file."
    .type = str
}
"""
)



from __future__ import annotations

import pytest

from cctbx import sgtbx
from cctbx.sgtbx.lattice_symmetry import metric_subgroups

from dials.algorithms.symmetry.cosym._generate_test_data import generate_intensities
from dials.algorithms.symmetry.laue_group import (
    ScoreCorrelationCoefficient,
    ScoreSubGroup,
    ScoreSymmetryElement,
)


def test_score_correlation_coefficient():
    cc = 1
    expected_cc = 1
    sigma_cc = 0.1
    score_cc = ScoreCorrelationCoefficient(cc, sigma_cc, expected_cc)
    assert score_cc.p_s_given_cc == pytest.approx(0.9492499653267421)

    cc = 0.5
    score_cc = ScoreCorrelationCoefficient(cc, sigma_cc, expected_cc)
    assert score_cc.p_s_given_cc == pytest.approx(0.19697339347266518)

    cc = 0.5
    expected_cc = 0.6
    sigma_cc = 0.2
    score_cc = ScoreCorrelationCoefficient(cc, sigma_cc, expected_cc)
    assert score_cc.p_s_given_cc == pytest.approx(0.6178439917879021)

    if 0:
        import numpy as np
        from matplotlib import pyplot as plt

        x = np.linspace(-1, 1)
        values = [(0.1, 1), (0.1, 0.8), (0.2, 1), (0.2, 0.8)]
        fig, axes = plt.subplots(nrows=2, ncols=2)
        for ax, (sigma_cc, expected_cc) in zip(axes.flatten(), values):
            ax.set_title(f"E(CC) = {expected_cc:.1f}, sigma(cc) = {sigma_cc:.1f}")
            p_cc_given_s = []
            p_cc_given_not_s = []
            p_s_given_cc = []
            for _ in x:
                score_cc = ScoreCorrelationCoefficient(_, sigma_cc, expected_cc)
                p_cc_given_s.append(score_cc.p_cc_given_s)
                p_cc_given_not_s.append(score_cc.p_cc_given_not_s)
                p_s_given_cc.append(score_cc.p_s_given_cc)
            ax.plot(x, p_cc_given_s, label="p(CC;S)")
            ax.plot(x, p_cc_given_not_s, label="p(CC;!S)")
            ax.plot(x, p_s_given_cc, label="p(S;CC)")
        ylim = max(ax.get_ylim()[1] for ax in axes.flatten())
        for ax in axes.flatten():
            ax.legend()
            ax.set_ylim(ax.get_ylim()[0], ylim)
        plt.show()


@pytest.mark.parametrize("space_group", ["P2", "P3", "P6", "R3:h", "I23"][:])
def test_score_symmetry_element_subgroup(space_group):
    sgi = sgtbx.space_group_info(symbol=space_group)
    cs = sgi.any_compatible_crystal_symmetry(volume=10000)
    cs = cs.best_cell()
    cs = cs.minimum_cell()
    intensities = (
        generate_intensities(cs, d_min=1.0)
        .generate_bijvoet_mates()
        .set_observation_type_xray_intensity()
    )
    intensities = intensities.expand_to_p1()

    subgroups = metric_subgroups(
        intensities.crystal_symmetry(), max_delta=2.0, bravais_types_only=False
    )
    intensities = intensities.change_basis(subgroups.cb_op_inp_minimum)
    cs = cs.change_basis(subgroups.cb_op_inp_minimum)

    cb_op_inp_best = subgroups.result_groups[0]["cb_op_inp_best"]
    lattice_group = subgroups.result_groups[0]["best_subsym"].space_group()
    lattice_group = lattice_group.change_basis(cb_op_inp_best.inverse())

    lattice_group = subgroups.result_groups[0]["subsym"].space_group()

    sym_op_scores = []
    for sym_op in lattice_group.smx():
        if sym_op.r().info().sense() < 0:
            continue
        score = ScoreSymmetryElement(intensities, sym_op, 1.0, 1.0)
        sym_op_scores.append(score)
        if sym_op in cs.space_group():
            assert score.likelihood > 0.9
            assert score.cc.coefficient() > 0.9
        else:
            assert score.likelihood < 0.2
            assert score.cc.coefficient() < 0.3

    subgroup_scores = [
        ScoreSubGroup(subgrp, sym_op_scores) for subgrp in subgroups.result_groups
    ]
    total_likelihood = sum(score.likelihood for score in subgroup_scores)
    for score in subgroup_scores:
        score.likelihood /= total_likelihood
    true_patterson_group = (
        cs.space_group_info()
        .as_reference_setting()
        .group()
        .build_derived_patterson_group()
    )
    for score in subgroup_scores:
        if score.subgroup["best_subsym"].space_group() == true_patterson_group:
            assert score.likelihood > 0.8
        else:
            assert score.likelihood < 0.1



from __future__ import annotations

from math import sqrt

import pytest

from scitbx import matrix


class Model:
    def __init__(self, test_nave_model=False):
        # Set up experimental models with regular geometry
        from dxtbx.model import BeamFactory, DetectorFactory, GoniometerFactory

        # Beam along the Z axis
        self.beam = BeamFactory.make_beam(unit_s0=matrix.col((0, 0, 1)), wavelength=1.0)

        # Goniometer (used only for index generation) along X axis
        self.goniometer = GoniometerFactory.known_axis(matrix.col((1, 0, 0)))

        # Detector fast, slow along X, -Y; beam in the centre, 200 mm distance
        dir1 = matrix.col((1, 0, 0))
        dir2 = matrix.col((0, -1, 0))
        centre = matrix.col((0, 0, 200))
        npx_fast = npx_slow = 1000
        pix_size = 0.2
        origin = centre - (
            0.5 * npx_fast * pix_size * dir1 + 0.5 * npx_slow * pix_size * dir2
        )
        self.detector = DetectorFactory.make_detector(
            "PAD",
            dir1,
            dir2,
            origin,
            (pix_size, pix_size),
            (npx_fast, npx_slow),
            (0, 1.0e6),
        )

        # Cubic 100 A^3 crystal
        a = matrix.col((100, 0, 0))
        b = matrix.col((0, 100, 0))
        c = matrix.col((0, 0, 100))

        if test_nave_model:
            from dxtbx.model import MosaicCrystalSauter2014

            self.crystal = MosaicCrystalSauter2014(a, b, c, space_group_symbol="P 1")
            self.crystal.set_half_mosaicity_deg(500)
            self.crystal.set_domain_size_ang(0.2)
        else:
            from dxtbx.model import Crystal

            self.crystal = Crystal(a, b, c, space_group_symbol="P 1")

        # Collect these models in an Experiment (ignoring the goniometer)
        from dxtbx.model.experiment_list import Experiment

        self.experiment = Experiment(
            beam=self.beam,
            detector=self.detector,
            goniometer=None,
            scan=None,
            crystal=self.crystal,
            imageset=None,
        )

        # Generate some reflections
        self.reflections = self.generate_reflections()

    def generate_reflections(self):
        """Use reeke_model to generate indices of reflections near to the Ewald
        sphere that might be observed on a still image. Build a reflection_table
        of these."""
        from cctbx.sgtbx import space_group_info

        space_group_type = space_group_info("P 1").group().type()

        # create a ReekeIndexGenerator
        UB = self.crystal.get_A()
        axis = self.goniometer.get_rotation_axis()
        s0 = self.beam.get_s0()
        # use the same UB at the beginning and end - the margin parameter ensures
        # we still have indices close to the Ewald sphere generated
        from dials.algorithms.spot_prediction import ReekeIndexGenerator

        r = ReekeIndexGenerator(UB, UB, space_group_type, axis, s0, dmin=1.5, margin=1)

        # generate indices
        hkl = r.to_array()
        nref = len(hkl)

        # create a reflection table
        from dials.array_family import flex

        table = flex.reflection_table()
        table["flags"] = flex.size_t(nref, 0)
        table["id"] = flex.int(nref, 0)
        table["panel"] = flex.size_t(nref, 0)
        table["miller_index"] = flex.miller_index(hkl)
        table["entering"] = flex.bool(nref, True)
        table["s1"] = flex.vec3_double(nref)
        table["xyzcal.mm"] = flex.vec3_double(nref)
        table["xyzcal.px"] = flex.vec3_double(nref)

        return table


@pytest.mark.parametrize("nave_model", [True, False], ids=["nave", "native"])
def test(nave_model):
    model = Model(test_nave_model=nave_model)

    # cache objects from the model
    UB = matrix.sqr(model.crystal.get_A())
    s0 = matrix.col(model.beam.get_s0())
    es_radius = s0.length()

    # create the predictor and predict for reflection table
    from dials.algorithms.spot_prediction import StillsReflectionPredictor

    predictor = StillsReflectionPredictor(model.experiment)
    predictor.for_reflection_table(model.reflections, UB)

    # for every reflection, reconstruct relp rotated to the Ewald sphere (vector
    # r) and unrotated relp (vector q), calculate the angle between them and
    # compare with delpsical.rad
    from libtbx.test_utils import approx_equal

    for ref in model.reflections.rows():
        r = matrix.col(ref["s1"]) - s0
        q = UB * matrix.col(ref["miller_index"])
        tst_radius = (s0 + q).length()
        sgn = -1 if tst_radius > es_radius else 1
        delpsi = sgn * r.accute_angle(q)
        assert approx_equal(delpsi, ref["delpsical.rad"])


def test_spherical_relps():
    model = Model()

    # cache objects from the model
    UB = matrix.sqr(model.crystal.get_A())
    s0 = matrix.col(model.beam.get_s0())
    es_radius = s0.length()

    # create the predictor and predict for reflection table
    from dials.algorithms.spot_prediction import StillsReflectionPredictor

    predictor = StillsReflectionPredictor(model.experiment, spherical_relp=True)
    predictor.for_reflection_table(model.reflections, UB)

    # for every reflection, reconstruct relp centre q, calculate s1 according
    # to the formula in stills_prediction_nave3.pdf and compare
    from libtbx.test_utils import approx_equal

    for ref in model.reflections.rows():
        q = UB * matrix.col(ref["miller_index"])
        radicand = q.length_sq() + 2.0 * q.dot(s0) + s0.length_sq()
        assert radicand > 0.0
        denom = sqrt(radicand)
        s1 = es_radius * (q + s0) / denom
        assert approx_equal(s1, ref["s1"])



from __future__ import annotations

import http.client
import json
import os
import select
import socket as pysocket
import sys
import urllib.error
import urllib.parse
import urllib.request
from multiprocessing.pool import ThreadPool

import libtbx.phil
from dxtbx.model.crystal import CrystalFactory
from dxtbx.util import get_url_scheme
from libtbx.introspection import number_of_processors
from scitbx.array_family import flex

import dials.util


def work(host, port, filename, params):
    conn = http.client.HTTPConnection(host, port)
    path = filename
    for param in params:
        path += f";{param}"
    conn.request("GET", path)
    return conn.getresponse().read()


def _nproc():
    return number_of_processors(return_value_if_unknown=-1)


def response_to_xml(d):

    if "n_spots_total" in d:
        response = f"""<image>{d['image']}</image>
<spot_count>{d['n_spots_total']}</spot_count>
<spot_count_no_ice>{d['n_spots_no_ice']}</spot_count_no_ice>
<d_min>{d['estimated_d_min']:.2f}</d_min>
<d_min_method_1>{d['d_min_distl_method_1']:.2f}</d_min_method_1>
<d_min_method_2>{d['d_min_distl_method_2']:.2f}</d_min_method_2>
<total_intensity>{d['total_intensity']:.0f}</total_intensity>"""

    else:
        assert "error" in d
        return f"<response>\n{d['error']}\n</response>"

    if "lattices" in d:

        for lattice in d["lattices"]:
            crystal = CrystalFactory.from_dict(lattice["crystal"])
            response = "\n".join(
                [
                    response,
                    "<unit_cell>%.6g %.6g %.6g %.6g %.6g %.6g</unit_cell>"
                    % (crystal.get_unit_cell().parameters()),
                ]
            )
        response = "\n".join(
            [
                response,
                "<n_indexed>%i</n_indexed>" % d["n_indexed"],
                "<fraction_indexed>%.2f</fraction_indexed>" % d["fraction_indexed"],
            ]
        )
    if "integrated_intensity" in d:
        response = "\n".join(
            [
                response,
                "<integrated_intensity>%.0f</integrated_intensity>"
                % d["integrated_intensity"],
            ]
        )

    return f"<response>\n{response}\n</response>"


def work_all(
    host,
    port,
    filenames,
    params,
    plot=False,
    table=False,
    json_file=None,
    grid=None,
    nproc=None,
):
    if nproc is None:
        nproc = _nproc()
    with ThreadPool(processes=nproc) as pool:
        threads = {}
        for filename in filenames:
            threads[filename] = pool.apply_async(work, (host, port, filename, params))
        results = []
        for filename in filenames:
            response = threads[filename].get()
            d = json.loads(response)
            results.append(d)
            print(response_to_xml(d))

    if json_file is not None:
        with open(json_file, "wb") as f:
            json.dump(results, f)

    if plot or table:

        from dials.algorithms.spot_finding.per_image_analysis import (
            StatsMultiImage,
            plot_stats,
        )

        estimated_d_min = flex.double()
        d_min_distl_method_1 = flex.double()
        d_min_distl_method_2 = flex.double()
        n_spots_total = flex.int()
        n_spots_no_ice = flex.int()
        total_intensity = flex.double()

        for d in results:
            estimated_d_min.append(d["estimated_d_min"])
            d_min_distl_method_1.append(d["d_min_distl_method_1"])
            d_min_distl_method_2.append(d["d_min_distl_method_2"])
            n_spots_total.append(d["n_spots_total"])
            n_spots_no_ice.append(d["n_spots_no_ice"])
            total_intensity.append(d["total_intensity"])

        stats = StatsMultiImage(
            n_spots_total=n_spots_total,
            n_spots_no_ice=n_spots_no_ice,
            n_spots_4A=None,
            total_intensity=total_intensity,
            estimated_d_min=estimated_d_min,
            d_min_distl_method_1=d_min_distl_method_1,
            d_min_distl_method_2=d_min_distl_method_2,
            noisiness_method_1=None,
            noisiness_method_2=None,
        )

        if plot:
            plot_stats(stats)
        if table:
            print(stats)

        if grid is not None:
            from matplotlib import pyplot

            n_spots_no_ice.reshape(flex.grid(grid))
            print(n_spots_no_ice.size())

            pyplot.figure()
            pyplot.pcolormesh(n_spots_no_ice.as_numpy_array(), cmap=pyplot.cm.Reds)
            pyplot.savefig("spot_count.png")


def stop(host, port, nproc):
    stopped = 0
    for j in range(nproc):
        try:
            url_request = urllib.request.Request(f"http://{host}:{port}/Ctrl-C")
            socket = urllib.request.urlopen(url_request, None, 3)
            if socket.getcode() == "200":
                stopped = stopped + 1
            else:
                print("socket returned code", socket.getcode())
        except (pysocket.timeout, urllib.error.HTTPError) as e:
            print("error on stopping server:", e)
        except urllib.error.URLError as e:
            if e.reason.errno != 111:
                print("error on stopping server:", e)
        except pysocket.error:
            # Assuming this means the server killed itself before the reply left the send buffer.
            stopped = stopped + 1
        except http.client.BadStatusLine:
            # Regular occurrence. Probably means the server stopped anyway.
            stopped = stopped + 1
    return stopped


phil_scope = libtbx.phil.parse(
    """\
nproc = Auto
  .type = int(value_min=1)
host = localhost
  .type = str
port = 1701
  .type = int(value_min=1)
plot = False
  .type = bool
table = False
  .type = bool
json = None
  .type = path
grid = None
  .type = ints(size=2, value_min=1)
"""
)


@dials.util.show_mail_handle_errors()
def run(args=None):
    mixed_args = args or sys.argv[1:]
    if os.name != "nt":
        r, w, x = select.select([sys.stdin], [], [], 0)
        if len(r) > 0:
            mixed_args.extend([l.strip() for rr in r for l in rr.readlines()])

    filenames = []
    args = []
    for arg in mixed_args:
        if get_url_scheme(arg):
            # Make this look like a path. If you squint. And are looking away.
            filenames.append("/" + urllib.parse.quote(arg))
        else:
            if os.path.isfile(arg):
                filenames.append(arg)
            else:
                args.append(arg)

    interp = phil_scope.command_line_argument_interpreter()
    params, unhandled = interp.process_and_fetch(
        args, custom_processor="collect_remaining"
    )
    params = params.extract()

    if params.nproc is libtbx.Auto:
        nproc = None
        params.nproc = 1024
    else:
        nproc = params.nproc

    if len(unhandled) and unhandled[0] == "stop":
        stopped = stop(params.host, params.port, params.nproc)
        print("Stopped %d findspots processes" % stopped)
    elif len(unhandled) and unhandled[0] == "ping":
        url = "http://%s:%i" % (params.host, params.port)
        try:
            _ = urllib.request.urlopen(url).read()
            print("Success")
            sys.exit(0)
        except Exception:
            print("Failure")
            sys.exit(1)
    else:
        if len(filenames) == 1:
            response = work(params.host, params.port, filenames[0], unhandled)

            print(response_to_xml(json.loads(response)))
        else:
            work_all(
                params.host,
                params.port,
                filenames,
                unhandled,
                plot=params.plot,
                table=params.table,
                json_file=params.json,
                grid=params.grid,
                nproc=nproc,
            )


if __name__ == "__main__":
    run()



from __future__ import annotations

import math
import random
import textwrap

import pytest

from dxtbx.model import Beam, Crystal, Goniometer
from libtbx.test_utils import approx_equal
from scitbx import matrix
from scitbx.array_family import flex

from dials.algorithms.refinement.parameterisation.scan_varying_beam_parameters import (
    ScanVaryingBeamParameterisation,
)
from dials.algorithms.refinement.parameterisation.scan_varying_crystal_parameters import (
    ScanVaryingCrystalOrientationParameterisation,
    ScanVaryingCrystalUnitCellParameterisation,
)
from dials.algorithms.refinement.parameterisation.scan_varying_detector_parameters import (
    ScanVaryingDetectorParameterisationSinglePanel,
)
from dials.algorithms.refinement.parameterisation.scan_varying_model_parameters import (
    GaussianSmoother,
    ScanVaryingParameterSet,
)
from dials.algorithms.refinement.refinement_helpers import (
    get_fd_gradients,
    random_param_shift,
)


def test_gaussian_smoother(plots=False):
    """Test a bare parameter set with the smoother"""

    # 7 values, all set to 1.0
    myparam = ScanVaryingParameterSet(1.0, 7)

    # Adjust a couple of the values
    myparam.value[3:4] = [2.0, 2.0]

    # Make a smoother with x_range as an 'image range', between 1 and 100.
    # This smoother needs 5 intervals (for 7 total values). The default
    # smoother uses an averaging window of 3 values
    smoother = GaussianSmoother((1, 100), 5)

    assert smoother.num_values() == 7
    assert smoother.num_samples() == 5
    assert smoother.num_average() == 3

    # The smoother positions depend on the number of intervals but not on
    # the x_range
    assert smoother.positions() == [-0.5, 0.5, 1.5, 2.5, 3.5, 4.5, 5.5]

    # By contrast the spacing is in units of the original, unnormalised
    # coordinate
    assert smoother.spacing() == 19.8

    # Use the single value smoother multiple times...
    smooth_at = list(range(1, 101))
    data = [smoother.value_weight(e, myparam) for e in smooth_at]
    vals, weights, sumweights = zip(*data)
    assert len(smooth_at) == len(vals)

    # ...and the multi value smoother once
    mvals, mweights, msumweights = smoother.multi_value_weight(smooth_at, myparam)

    # the results should be identical.
    assert (flex.double(vals) == mvals).all_eq(True)
    assert (flex.double(sumweights) == msumweights).all_eq(True)
    for v1, v2 in zip(weights, mweights.transpose().cols()):
        assert (v1.as_dense_vector() == v2.as_dense_vector()).all_eq(True)

    # make scatterplots
    if plots:
        import matplotlib.pyplot as plt

        plt.ion()
        plt.scatter(smooth_at, vals)
        plt.draw()


# Use classes to wrap up ScanVarying*Parameterisation classes with
# compose and get_state overridden, so they can be passed to the existing FD
# derivative code.
class _TestOrientationModel(ScanVaryingCrystalOrientationParameterisation):
    def __init__(self, image_number, *args):
        ScanVaryingCrystalOrientationParameterisation.__init__(self, *args)

        # set overloads now, after construction of the base class
        self.compose = self._compose
        self.get_state = self._get_state

        self.set_time_point(image_number)

    def set_time_point(self, t):
        self.image_number = t
        self.compose()

    def _compose(self):
        """override for compose to pass in the requested t"""

        ScanVaryingCrystalOrientationParameterisation.compose(self, self.image_number)

    def _get_state(self):
        """override for get_state to do so only at the requested t"""

        # ensure the state is updated by re-composing
        self.compose()
        return ScanVaryingCrystalOrientationParameterisation.get_state(self)


class _TestUnitCellModel(ScanVaryingCrystalUnitCellParameterisation):
    def __init__(self, image_number, *args):
        ScanVaryingCrystalUnitCellParameterisation.__init__(self, *args)

        # set overloads now, after construction of the base class
        self.compose = self._compose
        self.get_state = self._get_state

        self.set_time_point(image_number)

    def set_time_point(self, t):
        self.image_number = t
        self.compose()

    def _compose(self):
        """override for compose to pass in the requested t"""

        ScanVaryingCrystalUnitCellParameterisation.compose(self, self.image_number)

    def _get_state(self):
        """override for get_state to do so only at the requested t"""

        # ensure the state is updated by re-composing
        self.compose()
        return ScanVaryingCrystalUnitCellParameterisation.get_state(self)


class _TestBeamModel(ScanVaryingBeamParameterisation):
    def __init__(self, image_number, *args):
        ScanVaryingBeamParameterisation.__init__(self, *args)

        # set overloads now, after construction of the base class
        self.compose = self._compose
        self.get_state = self._get_state

        self.set_time_point(image_number)

    def set_time_point(self, t):
        self.image_number = t
        self.compose()

    def _compose(self):
        """override for compose to pass in the requested t"""

        ScanVaryingBeamParameterisation.compose(self, self.image_number)

    def _get_state(self):
        """override for get_state to do so only at the requested t"""

        # ensure the state is updated by re-composing
        self.compose()
        return ScanVaryingBeamParameterisation.get_state(self)


class _TestDetectorModel(ScanVaryingDetectorParameterisationSinglePanel):
    def __init__(self, image_number, *args):
        ScanVaryingDetectorParameterisationSinglePanel.__init__(self, *args)

        # set overloads now, after construction of the base class
        self.compose = self._compose
        self.get_state = self._get_state

        self.set_time_point(image_number)

    def set_time_point(self, t):
        self.image_number = t
        self.compose()

    def _compose(self):
        """override for compose to pass in the requested t"""

        ScanVaryingDetectorParameterisationSinglePanel.compose(self, self.image_number)

    def _get_state(self):
        """override for get_state to do so only at the requested t"""

        # ensure the state is updated by re-composing
        self.compose()
        return ScanVaryingDetectorParameterisationSinglePanel.get_state(self)


class _TestScanVaryingModelParameterisation:
    def __init__(self):
        # Let's say we have a scan of 100 images
        self.image_range = (1, 100)

        # Make a random P1 crystal
        a = random.uniform(10, 50) * self.random_direction_close_to(
            matrix.col((1, 0, 0))
        )
        b = random.uniform(10, 50) * self.random_direction_close_to(
            matrix.col((0, 1, 0))
        )
        c = random.uniform(10, 50) * self.random_direction_close_to(
            matrix.col((0, 0, 1))
        )
        self.xl = Crystal(a, b, c, space_group_symbol="P 1")

        # Make a beam with wavelength in the range 0.8--1.2 and s0 direction close
        # to 0,0,1
        s0 = random.uniform(0.8, 1.2) * self.random_direction_close_to(
            matrix.col((0, 0, 1))
        )
        self.beam = Beam(s0)

        # Make a standard goniometer model along X
        self.goniometer = Goniometer((1, 0, 0))

        # Make a simple single panel detector
        d1 = matrix.col((1, 0, 0))
        d2 = matrix.col((0, -1, 0))
        npx_fast = 1475
        npx_slow = 1679
        pix_size_f = pix_size_s = 0.172
        from dxtbx.model import DetectorFactory

        self.detector = DetectorFactory.make_detector(
            "PAD",
            d1,
            d2,
            matrix.col((0, 0, -110)),
            (pix_size_f, pix_size_s),
            (npx_fast, npx_slow),
            (0, 2e20),
        )

    def random_direction_close_to(self, vector):
        return vector.rotate_around_origin(
            matrix.col((random.random(), random.random(), random.random())).normalize(),
            random.gauss(0, 1.0),
            deg=True,
        )


@pytest.mark.parametrize("nintervals", (1, 2, 3, 4, 5, 6, 7))
def test_ScanVaryingCrystalOrientationParameterisation_intervals(
    nintervals, plots=False
):
    """Test a ScanVaryingCrystalOrientationParameterisation with
    a range of different numbers of intervals"""

    vmp = _TestScanVaryingModelParameterisation()

    # Parameterise the crystal with the image range and five intervals. Init
    # TestOrientationModel to explore gradients at image 50, but actually
    # will try various time points in the test
    xl_op = _TestOrientationModel(50, vmp.xl, vmp.image_range, nintervals)

    # How many parameters?
    num_param = xl_op.num_free()

    # shift the parameters away from zero
    p_vals = xl_op.get_param_vals()
    sigmas = [1.0] * len(p_vals)
    new_vals = random_param_shift(p_vals, sigmas)
    xl_op.set_param_vals(new_vals)

    # recalc state and gradients at image 50
    xl_op.compose()
    p_vals = xl_op.get_param_vals()
    # print "Shifted parameter vals", p_vals

    # compare analytical and finite difference derivatives at image 50
    an_ds_dp = xl_op.get_ds_dp()
    fd_ds_dp = get_fd_gradients(xl_op, [1.0e-6 * math.pi / 180] * num_param)

    null_mat = matrix.sqr((0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0))
    for e, f in zip(an_ds_dp, fd_ds_dp):
        assert approx_equal((e - f), null_mat, eps=1.0e-6)

    # Now test gradients at equally spaced time points across the whole
    # range
    num_points = 50
    smooth_at = []
    phi1_data = []
    phi2_data = []
    phi3_data = []
    step_size = (vmp.image_range[1] - vmp.image_range[0]) / num_points
    for t in [vmp.image_range[0] + e * step_size for e in range(num_points + 1)]:

        # collect data for plot
        smooth_at.append(t)
        phi1_data.append(xl_op._smoother.value_weight(t, xl_op._param[0])[0])
        phi2_data.append(xl_op._smoother.value_weight(t, xl_op._param[1])[0])
        phi3_data.append(xl_op._smoother.value_weight(t, xl_op._param[2])[0])

        xl_op.set_time_point(t)
        an_ds_dp = xl_op.get_ds_dp()
        fd_ds_dp = get_fd_gradients(xl_op, [1.0e-6 * math.pi / 180] * num_param)
        for e, f in zip(an_ds_dp, fd_ds_dp):
            assert approx_equal((e - f), null_mat, eps=1.0e-6)

    if plots:
        import matplotlib.pyplot as plt

        plt.ion()
        plt.clf()
        plt.subplot(311)
        plt.cla()
        plt.scatter(smooth_at, phi1_data)
        plt.title("Phi1")
        plt.xlabel("image number")
        plt.ylabel("Phi1 (mrad)")
        plt.subplot(312)
        plt.cla()
        plt.scatter(smooth_at, phi2_data)
        plt.title("Phi2")
        plt.xlabel("image number")
        plt.ylabel("Phi2 (mrad)")
        plt.subplot(313)
        plt.cla()
        plt.scatter(smooth_at, phi3_data)
        plt.title("Phi3")
        plt.xlabel("image number")
        plt.ylabel("Phi3 (mrad)")
        plt.suptitle("Parameter smoothing with %d intervals" % nintervals)
        plt.draw()


def test_ScanVaryingCrystalOrientationParameterisation_random(plots=False):
    """Test a ScanVaryingCrystalOrientationParameterisation with
    random initial orientations, random parameter shifts and random times"""

    vmp = _TestScanVaryingModelParameterisation()

    attempts = 100
    null_mat = matrix.sqr((0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0))

    for i in range(attempts):
        # make a new P1 random crystal and parameterise it
        a = random.uniform(10, 50) * vmp.random_direction_close_to(
            matrix.col((1, 0, 0))
        )
        b = random.uniform(10, 50) * vmp.random_direction_close_to(
            matrix.col((0, 1, 0))
        )
        c = random.uniform(10, 50) * vmp.random_direction_close_to(
            matrix.col((0, 0, 1))
        )
        xl = Crystal(a, b, c, space_group_symbol="P 1")

        xl_op = _TestOrientationModel(50, xl, vmp.image_range, 5)

        # How many parameters?
        num_param = xl_op.num_free()

        # apply random parameter shifts to the orientation (2.0 mrad each
        # checkpoint)
        p_vals = xl_op.get_param_vals()
        sigmas = [2.0] * len(p_vals)
        new_vals = random_param_shift(p_vals, sigmas)
        xl_op.set_param_vals(new_vals)

        # select random time point at which to make comparisons
        t = random.uniform(*vmp.image_range)
        xl_op.set_time_point(t)

        # compare analytical and finite difference derivatives
        xl_op_an_ds_dp = xl_op.get_ds_dp()
        xl_op_fd_ds_dp = get_fd_gradients(xl_op, [1.0e-6 * math.pi / 180] * num_param)

        for j in range(num_param):
            assert approx_equal(
                (xl_op_fd_ds_dp[j] - xl_op_an_ds_dp[j]), null_mat, eps=1.0e-6
            ), textwrap.dedent(
                """\
        Failure in try {i}
        failure for parameter number {j}
        of the orientation parameterisation
        with fd_ds_dp =
        {fd}
        and an_ds_dp =
        {an}
        so that difference fd_ds_dp - an_ds_dp =
        {diff}
        """
            ).format(
                i=i,
                j=j,
                fd=xl_op_fd_ds_dp[j],
                an=xl_op_an_ds_dp[j],
                diff=xl_op_fd_ds_dp[j] - xl_op_an_ds_dp[j],
            )


@pytest.mark.parametrize("nintervals", (1, 2, 3, 4, 5, 6, 7))
def test_ScanVaryingCrystalUnitCellParameterisation_intervals(nintervals, plots=False):
    """Basic test of a ScanVaryingCrystalUnitCellParameterisation
    with a range of different numbers of intervals"""

    vmp = _TestScanVaryingModelParameterisation()

    # Parameterise the crystal with the image range and five intervals. Init
    # TestOrientationModel to explore gradients at image 50
    xl_ucp = _TestUnitCellModel(50, vmp.xl, vmp.image_range, nintervals)

    # How many parameters?
    num_param = xl_ucp.num_free()

    # apply a random parameter shift to the unit cell, on order of 2% of
    # the initial metrical matrix parameters
    p_vals = xl_ucp.get_param_vals()
    sigmas = [0.02 * p for p in p_vals]
    new_vals = random_param_shift(p_vals, sigmas)
    xl_ucp.set_param_vals(new_vals)

    # calculate state and gradients at image 50
    xl_ucp.compose()

    null_mat = matrix.sqr((0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0))

    # compare analytical and finite difference derivatives at image 50
    an_ds_dp = xl_ucp.get_ds_dp()
    fd_ds_dp = get_fd_gradients(xl_ucp, [1.0e-7] * num_param)

    for e, f in zip(an_ds_dp, fd_ds_dp):
        assert approx_equal((e - f), null_mat, eps=1.0e-6)


@pytest.mark.parametrize("nintervals", (1, 2, 3, 4, 5, 6, 7))
def test_ScanVaryingBeamParameterisation(nintervals, plots=False):
    """Basic test of a ScanVaryingBeamParameterisation
    with a range of different numbers of intervals"""

    vmp = _TestScanVaryingModelParameterisation()

    # Parameterise the crystal with the image range and five intervals. Init
    # TestOrientationModel to explore gradients at image 50
    beam_p = _TestBeamModel(50, vmp.beam, vmp.image_range, nintervals, vmp.goniometer)

    # How many parameters?
    num_param = beam_p.num_free()

    # apply a random parameter shift to the beam, on order of 2% of
    # the initial values
    p_vals = beam_p.get_param_vals()
    sigmas = [0.02 * p for p in p_vals]
    new_vals = random_param_shift(p_vals, sigmas)
    beam_p.set_param_vals(new_vals)

    # calculate state and gradients at image 50
    beam_p.compose()

    null_vec = matrix.col((0.0, 0.0, 0.0))

    # compare analytical and finite difference derivatives at image 50
    an_ds_dp = beam_p.get_ds_dp()
    fd_ds_dp = get_fd_gradients(beam_p, [1.0e-7] * num_param)

    for e, f in zip(an_ds_dp, fd_ds_dp):
        assert approx_equal((e - f), null_vec, eps=1.0e-6)


@pytest.mark.parametrize("nintervals", (1, 2, 3, 4, 5, 6, 7))
def test_ScanVaryingDetectorParameterisation(nintervals, plots=False):
    """Basic test of a ScanVaryingDetectorParameterisationSinglePanel
    with a range of different numbers of intervals"""

    vmp = _TestScanVaryingModelParameterisation()

    # Parameterise the detector with the image range and five intervals. Init
    # TestOrientationModel to explore gradients at image 50
    det_p = _TestDetectorModel(50, vmp.detector, vmp.image_range, nintervals)

    # How many parameters?
    num_param = det_p.num_free()

    # apply a random parameter shift to the detector, on order of 2% of
    # the initial values
    p_vals = det_p.get_param_vals()
    sigmas = [0.02 * p for p in p_vals]
    new_vals = random_param_shift(p_vals, sigmas)
    det_p.set_param_vals(new_vals)

    # calculate state and gradients at image 50
    det_p.compose()

    null_mat = matrix.sqr((0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0))

    # compare analytical and finite difference derivatives at image 50
    an_ds_dp = det_p.get_ds_dp()
    fd_ds_dp = get_fd_gradients(det_p, [1.0e-7] * num_param)

    for e, f in zip(an_ds_dp, fd_ds_dp):
        assert approx_equal((e - f), null_mat, eps=1.0e-6)



from __future__ import annotations

from pathlib import Path

import procrunner

from dials.algorithms.refinement.engine import Journal


def test_scan_varying_refinement_of_a_multiple_panel_detector(
    dials_regression, tmp_path
):
    from dials.array_family import flex

    data = Path(dials_regression) / "refinement_test_data" / "i23_as_24_panel_barrel"

    result = procrunner.run(
        [
            "dials.refine",
            data / "experiments.json",
            data / "indexed.pickle",
            "scan_varying=true",
            "history=history.json",
            "outlier.separate_blocks=False",
        ],
        working_directory=tmp_path,
    )
    assert not result.returncode and not result.stderr

    # there are plenty of things we could do with the refinement history, but
    # here just check that final RMSDs are low enough
    history = Journal.from_json_file(tmp_path / "history.json")
    final_rmsd = history["rmsd"][-1]
    assert final_rmsd[0] < 0.05
    assert final_rmsd[1] < 0.04
    assert final_rmsd[2] < 0.0002

    # also check that the used_in_refinement flag got set correctly
    rt = flex.reflection_table.from_file(tmp_path / "refined.refl")
    uir = rt.get_flags(rt.flags.used_in_refinement)
    assert uir.count(True) == history["num_reflections"][-1]



from __future__ import annotations


def get_inverse_ub_matrix_from_xparm(handle):
    """Get the inverse_ub_matrix from an xparm file handle

    Params:
        handle The file handle

    Returns:
        The inverse_ub_matrix
    """
    from scitbx import matrix

    return matrix.sqr(
        handle.unit_cell_a_axis + handle.unit_cell_b_axis + handle.unit_cell_c_axis
    )


def get_space_group_type_from_xparm(handle):
    """Get the space group tyoe object from an xparm file handle

    Params:
        handle The file handle

    Returns:
        The space group type object
    """
    from cctbx import sgtbx

    return sgtbx.space_group_type(
        sgtbx.space_group(sgtbx.space_group_symbols(handle.space_group).hall())
    )



from __future__ import annotations

import procrunner

from dials.array_family import flex


def test_spots_xds(tmp_path):
    xds_input = "SPOT.XDS"
    output_pickle = "spot.refl"

    tmp_path.joinpath(xds_input).write_text(
        """\
 2411.40 1000.70 25.00 16384. 0 0 0
 1328.60 2170.40 20.57 7326. 0 0 0
 177.56 2191.30 24.94 6779. 0 0 0
 1231.34 1220.04 24.99 1952. 0 0 0
 1227.07 1230.56 24.81 349. 0 0 0
 1341.63 1243.25 5.64 321. 2 -2 11
 1125.23 1197.72 12.14 231. -1 2 -10
 1317.52 1171.59 19.28 120. 6 -4 6
 1260.25 1300.55 13.67 116. -4 2 6
 1090.27 1199.47 41.49 114. -2 3 -13
"""
    )

    result = procrunner.run(
        [
            "dials.import_xds",
            xds_input,  # xparm_file,
            "input.method=reflections",
            "output.filename=" + output_pickle,
            "remove_invalid=True",
        ],
        working_directory=tmp_path,
    )
    assert not result.returncode and not result.stderr
    assert tmp_path.joinpath(output_pickle).is_file()

    reflections = flex.reflection_table.from_file(tmp_path / output_pickle)
    assert len(reflections) == 5

    tmp_path.joinpath(xds_input).unlink()
    assert not tmp_path.joinpath(xds_input).exists()

    # now test we can export it again
    result = procrunner.run(
        ["dials.export", "format=xds", output_pickle], working_directory=tmp_path
    )
    assert not result.returncode and not result.stderr
    assert tmp_path.joinpath("xds", "SPOT.XDS").is_file()

    txt = tmp_path.joinpath("xds", "SPOT.XDS").read_text()
    assert [line.strip() for line in txt.split("\n")] == [
        line.strip()
        for line in """\
 1341.63 1243.25 5.64 321.00  2 -2 11
 1125.23 1197.72 12.14 231.00  -1 2 -10
 1317.52 1171.59 19.28 120.00  6 -4 6
 1260.25 1300.55 13.67 116.00  -4 2 6
 1090.27 1199.47 41.49 114.00  -2 3 -13
""".split(
            "\n"
        )
    ]


def test_export_xds(dials_data, tmp_path):
    experiment = dials_data("centroid_test_data", pathlib=True) / "experiments.json"
    result = procrunner.run(
        ["dials.find_spots", "nproc=1", experiment], working_directory=tmp_path
    )
    assert not result.returncode and not result.stderr
    assert tmp_path.joinpath("strong.refl").is_file()

    result = procrunner.run(
        [
            "dials.export",
            "format=xds",
            experiment,
            "strong.refl",
        ],
        working_directory=tmp_path,
    )
    assert not result.returncode and not result.stderr
    assert tmp_path.joinpath("xds", "XDS.INP").is_file()
    assert tmp_path.joinpath("xds", "XPARM.XDS").is_file()
    assert tmp_path.joinpath("xds", "SPOT.XDS").is_file()

    tmp_path.joinpath("xds", "XDS.INP").unlink()
    tmp_path.joinpath("xds", "XPARM.XDS").unlink()
    assert not tmp_path.joinpath("xds", "XDS.INP").is_file()
    assert not tmp_path.joinpath("xds", "XPARM.XDS").is_file()
    result = procrunner.run(
        [
            "dials.export",
            "format=xds",
            experiment,
        ],
        working_directory=tmp_path,
    )
    assert not result.returncode and not result.stderr
    assert tmp_path.joinpath("xds", "XDS.INP").is_file()
    assert tmp_path.joinpath("xds", "XPARM.XDS").is_file()


def test_export_imported_experiments(dials_data, tmp_path):
    result = procrunner.run(
        [
            "dials.export",
            "format=xds",
            dials_data("centroid_test_data", pathlib=True)
            / "imported_experiments.json",
        ],
        working_directory=tmp_path,
    )
    assert not result.returncode and not result.stderr
    assert tmp_path.joinpath("xds", "XDS.INP").is_file()
    assert not tmp_path.joinpath("xds", "XPARM.XDS").is_file()
    assert not tmp_path.joinpath("xds", "SPOT.XDS").is_file()



# test running data reduction programs on ssx data
from __future__ import annotations

import procrunner


def test_ssx_reduction(dials_data, tmp_path):
    """
    Check that dials.cosym, dials.scale, dials.export and dials.merge run
    successfully on ssx data.
    Also test a few smaller analysis programs.
    """
    ssx = dials_data("cunir_serial_processed", pathlib=True)
    ssx_data = dials_data("cunir_serial", pathlib=True)
    refls = ssx / "integrated.refl"
    expts = ssx / "integrated.expt"

    result = procrunner.run(
        ["dials.cosym", expts, refls],
        working_directory=tmp_path,
    )
    assert not result.returncode and not result.stderr
    cosym_expts = tmp_path / "symmetrized.expt"
    cosym_refls = tmp_path / "symmetrized.refl"
    assert cosym_expts.is_file()
    assert cosym_refls.is_file()
    assert (tmp_path / "dials.cosym.html").is_file()

    result = procrunner.run(
        ["dials.scale", cosym_expts, cosym_refls],
        working_directory=tmp_path,
    )
    assert not result.returncode and not result.stderr
    scale_expts = tmp_path / "scaled.expt"
    scale_refls = tmp_path / "scaled.refl"
    assert scale_expts.is_file()
    assert scale_refls.is_file()
    assert (tmp_path / "dials.scale.html").is_file()

    # run scaling with reference model / cif
    for reference in [
        ssx_data / "2BW4.pdb",
        ssx_data / "2bw4.cif",
        ssx_data / "2bw4-sf.cif",
    ]:
        result = procrunner.run(
            [
                "dials.scale",
                cosym_expts,
                cosym_refls,
                f"reference={reference}",
                "output.experiments=scaled_ref.expt",
                "output.reflections=scaled_ref.refl",
            ],
            working_directory=tmp_path,
        )
    assert not result.returncode and not result.stderr

    result = procrunner.run(
        ["dials.export", scale_expts, scale_refls],
        working_directory=tmp_path,
    )
    assert not result.returncode and not result.stderr
    assert (tmp_path / "scaled.mtz").is_file()

    result = procrunner.run(
        ["dials.merge", scale_expts, scale_refls],
        working_directory=tmp_path,
    )
    assert not result.returncode and not result.stderr
    assert (tmp_path / "merged.mtz").is_file()

    result = procrunner.run(
        ["dials.damage_analysis", scale_expts, scale_refls],
        working_directory=tmp_path,
    )
    assert not result.returncode and not result.stderr
    assert (tmp_path / "dials.damage_analysis.html").is_file()

    result = procrunner.run(
        ["dials.compute_delta_cchalf", scale_expts, scale_refls],
        working_directory=tmp_path,
    )
    assert not result.returncode and not result.stderr
    assert (tmp_path / "compute_delta_cchalf.html").is_file()



from __future__ import annotations

import pytest

import libtbx
from cctbx import sgtbx

from dials.algorithms.symmetry.cosym import CosymAnalysis, phil_scope
from dials.algorithms.symmetry.cosym._generate_test_data import generate_test_data


@pytest.mark.parametrize(
    (
        "space_group",
        "unit_cell",
        "dimensions",
        "sample_size",
        "use_known_space_group",
        "use_known_lattice_group",
        "best_monoclinic_beta",
    ),
    [
        ("P2", None, None, 10, False, False, True),
        ("P3", None, None, 20, False, False, True),
        ("I23", None, libtbx.Auto, 10, False, False, True),
        ("I23", None, libtbx.Auto, 10, True, False, True),
        ("P422", (79, 79, 37, 90, 90, 90), None, 20, True, True, True),
        ("P321", (59.39, 59.39, 28.35, 90, 90, 120), None, 5, False, False, True),
        (
            "C121",
            (112.67, 52.85, 44.47, 90.00, 102.97, 90.00),
            None,
            5,
            False,
            False,
            False,
        ),
        (
            "C121",
            (112.67, 52.85, 44.47, 90.00, 102.97, 90.00),
            None,
            5,
            True,
            True,
            False,
        ),
        (
            "I121",
            (44.47, 52.85, 111.46, 90.00, 99.91, 90.00),
            None,
            5,
            False,
            False,
            True,
        ),
        (
            "I121",
            (44.47, 52.85, 111.46, 90.00, 99.91, 90.00),
            None,
            5,
            True,
            True,
            True,
        ),
    ],
)
def test_cosym(
    space_group,
    unit_cell,
    dimensions,
    sample_size,
    use_known_space_group,
    use_known_lattice_group,
    best_monoclinic_beta,
    tmp_path,
):
    import matplotlib

    matplotlib.use("Agg")

    datasets, expected_reindexing_ops = generate_test_data(
        space_group=sgtbx.space_group_info(symbol=space_group).group(),
        unit_cell=unit_cell,
        unit_cell_volume=10000,
        d_min=1.5,
        map_to_p1=True,
        sample_size=sample_size,
        seed=1,
    )
    expected_space_group = sgtbx.space_group_info(symbol=space_group).group()

    params = phil_scope.extract()
    params.dimensions = dimensions
    params.best_monoclinic_beta = best_monoclinic_beta
    if use_known_space_group:
        params.space_group = expected_space_group.info()
    if use_known_lattice_group:
        params.lattice_group = expected_space_group.info()

    params.normalisation = None
    cosym = CosymAnalysis(datasets, params)
    cosym.run()
    d = cosym.as_dict()
    if not use_known_space_group:
        assert d["subgroup_scores"][0]["likelihood"] > 0.89
        assert (
            sgtbx.space_group(d["subgroup_scores"][0]["patterson_group"])
            == sgtbx.space_group_info(space_group)
            .group()
            .build_derived_patterson_group()
        )
        expected_sg = (
            sgtbx.space_group_info(space_group).group().build_derived_patterson_group()
        )
    else:
        expected_sg = sgtbx.space_group_info(space_group).group()
    assert cosym.best_subgroup["best_subsym"].space_group() == expected_sg
    assert len(cosym.reindexing_ops) == len(expected_reindexing_ops)

    space_group_info = cosym.best_subgroup["subsym"].space_group_info()
    reference = None
    for d_id, cb_op in enumerate(cosym.reindexing_ops):
        reindexed = (
            datasets[d_id]
            .change_basis(sgtbx.change_of_basis_op(cb_op))
            .customized_copy(
                space_group_info=space_group_info.change_basis(
                    cosym.cb_op_inp_min.inverse()
                )
            )
        )
        assert reindexed.is_compatible_unit_cell(), str(reindexed.crystal_symmetry())
        if reference:
            assert (
                reindexed.correlation(
                    reference, assert_is_similar_symmetry=False
                ).coefficient()
                > 0.99
            )
        else:
            reference = reindexed



from __future__ import annotations

import pytest

from dxtbx.model import ExperimentList

from dials.array_family import flex
from dials.util import detect_blanks


def test_blank_counts_analysis(dials_data):
    expts = ExperimentList.from_file(
        dials_data("insulin_processed", pathlib=True) / "imported.expt",
        check_format=False,
    )
    refl = flex.reflection_table.from_file(
        dials_data("insulin_processed", pathlib=True) / "strong.refl"
    )
    results = detect_blanks.blank_counts_analysis(
        refl, expts[0].scan, phi_step=5, fractional_loss=0.1
    )
    assert set(results) == {"data", "layout", "blank_regions"}
    assert results["data"][0]["x"] == [
        2.5,
        7.5,
        12.5,
        17.5,
        22.5,
        27.5,
        32.5,
        37.5,
        42.5,
    ]
    assert results["data"][0]["y"] == [
        2827,
        2589,
        2502,
        2464,
        2515,
        2490,
        2441,
        2500,
        2505,
    ]
    assert not any(results["data"][0]["blank"])
    assert results["blank_regions"] == []

    z = refl["xyzobs.px.value"].parts()[2]
    refl_subset = refl.select((z < 10) | (z > 20))
    results = detect_blanks.blank_counts_analysis(
        refl_subset, expts[0].scan, phi_step=2, fractional_loss=0.1
    )
    assert results["data"][0]["blank"].count(True) == 5
    assert results["blank_regions"] == [(10, 21)]


def test_blank_integrated_analysis(dials_data):
    expts = ExperimentList.from_file(
        dials_data("insulin_processed", pathlib=True) / "integrated.expt",
        check_format=False,
    )
    refl = flex.reflection_table.from_file(
        dials_data("insulin_processed", pathlib=True) / "integrated.refl"
    )
    results = detect_blanks.blank_integrated_analysis(
        refl, expts[0].scan, phi_step=5, fractional_loss=0.1
    )
    assert results["data"][0]["x"] == [
        2.5,
        7.5,
        12.5,
        17.5,
        22.5,
        27.5,
        32.5,
        37.5,
        42.5,
    ]
    assert results["data"][0]["y"] == pytest.approx(
        [
            27.903266149430973,
            25.832527090455052,
            26.9236206883069,
            26.50234804728626,
            26.41019377727383,
            25.810676090828185,
            24.844906790823064,
            25.89992001081651,
            25.580718362291474,
        ]
    )
    assert not any(results["data"][0]["blank"])
    assert results["blank_regions"] == []

    # Now with some "blank" regions - make some of the reflections weak
    z = refl["xyzobs.px.value"].parts()[2]
    refl["intensity.prf.value"].set_selected(z < 10, refl["intensity.prf.value"] * 0.05)
    results = detect_blanks.blank_integrated_analysis(
        refl, expts[0].scan, phi_step=5, fractional_loss=0.1
    )
    assert results["data"][0]["y"] == pytest.approx(
        [
            1.3951633074715482,
            1.2916263545227527,
            26.9236206883069,
            26.50234804728626,
            26.41019377727383,
            25.810676090828185,
            24.844906790823064,
            25.89992001081651,
            25.580718362291474,
        ]
    )
    assert results["data"][0]["blank"] == [
        True,
        True,
        False,
        False,
        False,
        False,
        False,
        False,
        False,
    ]
    assert results["blank_regions"] == [(0, 10)]

    # Unset the integrated_prf flags, so the analysis should instead use the umodified
    # intensity.sum.value instead
    refl.unset_flags(flex.bool(len(refl), True), refl.flags.integrated_prf)
    results = detect_blanks.blank_integrated_analysis(
        refl, expts[0].scan, phi_step=5, fractional_loss=0.1
    )
    assert not any(results["data"][0]["blank"])
    assert results["blank_regions"] == []



from __future__ import annotations

from dials.algorithms.refinement.parameterisation.prediction_parameters import (
    PredictionParameterisation,
    SparseGradientVectorMixin,
)
from dials.array_family import flex
from dials_refinement_helpers_ext import dRq_de


class StillsPredictionParameterisation(PredictionParameterisation):
    """Concrete class that inherits functionality of the
    PredictionParameterisation parent class and provides a detector space
    implementation of the get_gradients function for still images. Gradients of
    the minimum rotation to the Ewald sphere, DeltaPsi, are calculated for use
    as a restraint in refinement"""

    _grad_names = ("dX_dp", "dY_dp", "dDeltaPsi_dp")

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # check that a goniometer parameterisation is not passed in
        assert not self._goniometer_parameterisations
        return

    def _local_setup(self, reflections):
        """Setup additional attributes used in gradients calculation. These are
        specific to the stills-type prediction parameterisations that rotates the
        relp q onto the Ewald sphere by an angle DeltaPsi before prediction"""

        self._DeltaPsi = reflections["delpsical.rad"]

        # q is the reciprocal lattice vector, in the lab frame
        self._q = self._UB * self._h
        self._q_scalar = self._q.norms()
        self._qq = self._q_scalar * self._q_scalar

        # r is the reciprocal lattice vector rotated to the Ewald sphere
        self._r = self._s1 - self._s0

        # we also need the unit directions q0 and s0u
        self._q0 = self._q.each_normalize()
        self._s0u = self._s0.each_normalize()

        # e1 is the unit vector about which DeltaPsi rotation is defined
        self._e1 = self._q0.cross(self._s0u).each_normalize()

        # q1 completes an orthonormal set with q0 and e1
        self._q1 = self._q0.cross(self._e1).each_normalize()

        # c0 completes an orthonormal set with s0u and e1
        self._c0 = self._s0u.cross(self._e1).each_normalize()

        # we want the wavelength
        self._wavelength = 1.0 / self._s0.norms()

        return

    def _beam_derivatives(self, isel, parameterisation=None, reflections=None):
        """helper function to extend the derivatives lists by derivatives of the
        beam parameterisations"""

        # Get required data
        s0 = self._s0.select(isel)
        s0u = self._s0u.select(isel)
        wl = self._wavelength.select(isel)
        r = self._r.select(isel)
        e1 = self._e1.select(isel)
        q = self._q.select(isel)
        c0 = self._c0.select(isel)
        DeltaPsi = self._DeltaPsi.select(isel)
        D = self._D.select(isel)

        # get the derivatives of the beam vector wrt the parameters
        ds0_dbeam_p = parameterisation.get_ds_dp(use_none_as_null=True)

        dDeltaPsi_dp = []
        dpv_dp = []

        # loop through the parameters
        for der in ds0_dbeam_p:

            if der is None:
                dpv_dp.append(None)
                dDeltaPsi_dp.append(None)
                continue

            # repeat the derivative in an array
            ds0 = flex.vec3_double(len(s0u), der.elems)

            # we need the derivative of the unit beam direction too. This requires
            # scaling by the wavelength and projection onto the Ewald sphere
            scaled = ds0 * wl
            ds0u = scaled.dot(c0) * c0 + scaled.dot(e1) * e1

            # calculate the derivative of DeltaPsi for this parameter
            dDeltaPsi = -1.0 * (r.dot(ds0)) / (e1.cross(r).dot(s0))
            dDeltaPsi_dp.append(dDeltaPsi)

            # calculate (d[r]/d[e1])(d[e1]/dp)
            de1_dp = c0.cross(ds0u)
            dr_de1 = dRq_de(DeltaPsi, e1, q)
            drde_dedp = dr_de1 * de1_dp

            # dp = 1.e-8 # finite step size for the parameter
            # del_e1 = de1_dp * dp
            # e1f = e1 + del_e1 * 0.5
            # rfwd = q.rotate_around_origin(e1f, DeltaPsi)
            # e1r = e1 - del_e1 * 0.5
            # rrev = q.rotate_around_origin(e1r, DeltaPsi)
            # drde_dedp = (rfwd - rrev) * (1 / dp)

            # calculate the derivative of pv for this parameter
            dpv_dp.append(D * (ds0 + e1.cross(r) * dDeltaPsi + drde_dedp))

        return dpv_dp, dDeltaPsi_dp

    def _xl_orientation_derivatives(
        self, isel, parameterisation=None, reflections=None
    ):
        """helper function to extend the derivatives lists by derivatives of the
        crystal orientation parameterisations"""

        # Get required data
        B = self._B.select(isel)
        h = self._h.select(isel)
        e1 = self._e1.select(isel)
        DeltaPsi = self._DeltaPsi.select(isel)
        s1 = self._s1.select(isel)
        q = self._q.select(isel)
        q_scalar = self._q_scalar.select(isel)
        qq = self._qq.select(isel)
        q0 = self._q0.select(isel)
        r = self._r.select(isel)
        s0 = self._s0.select(isel)
        s0u = self._s0u.select(isel)
        D = self._D.select(isel)

        # get derivatives of the U matrix wrt the parameters
        dU_dxlo_p = parameterisation.get_ds_dp(use_none_as_null=True)

        dDeltaPsi_dp = []
        dpv_dp = []

        # loop through the parameters
        for der in dU_dxlo_p:

            if der is None:
                dpv_dp.append(None)
                dDeltaPsi_dp.append(None)
                continue

            der_mat = flex.mat3_double(len(B), der.elems)

            # calculate the derivative of q for this parameter
            dq = der_mat * B * h

            # calculate the derivative of r for this parameter
            dr = dq.rotate_around_origin(e1, DeltaPsi)

            # calculate the derivative of DeltaPsi for this parameter
            dDeltaPsi = -1.0 * (dr.dot(s1)) / (e1.cross(r).dot(s0))
            dDeltaPsi_dp.append(dDeltaPsi)

            # derivative of the axis e1
            q_dot_dq = q.dot(dq)
            dq0 = (q_scalar * dq - (q_dot_dq * q0)) / qq
            de1_dp = dq0.cross(s0u)

            # calculate (d[r]/d[e1])(d[e1]/dp)
            # dr_de1 = dRq_de(DeltaPsi, e1, q)
            # drde_dedp = dr_de1 * de1_dp

            # The above calculation is not correct. Instead, calculate the partial
            # derivative of r wrt change in rotation axis e1 by finite differences
            dp = 1.0e-8
            del_e1 = de1_dp * dp
            e1f = e1 + del_e1 * 0.5
            rfwd = q.rotate_around_origin(e1f, DeltaPsi)
            e1r = e1 - del_e1 * 0.5
            rrev = q.rotate_around_origin(e1r, DeltaPsi)
            drde_dedp = (rfwd - rrev) * (1 / dp)

            # calculate the derivative of pv for this parameter
            dpv_dp.append(D * (dr + e1.cross(r) * dDeltaPsi + drde_dedp))

        return dpv_dp, dDeltaPsi_dp

    def _xl_unit_cell_derivatives(self, isel, parameterisation=None, reflections=None):
        """helper function to extend the derivatives lists by derivatives of the
        crystal unit cell parameterisations"""

        # Get required data
        U = self._U.select(isel)
        h = self._h.select(isel)
        e1 = self._e1.select(isel)
        DeltaPsi = self._DeltaPsi.select(isel)
        s1 = self._s1.select(isel)
        q = self._q.select(isel)
        q_scalar = self._q_scalar.select(isel)
        qq = self._qq.select(isel)
        q0 = self._q0.select(isel)
        r = self._r.select(isel)
        s0 = self._s0.select(isel)
        s0u = self._s0u.select(isel)
        D = self._D.select(isel)

        # get derivatives of the B matrix wrt the parameters
        dB_dxluc_p = parameterisation.get_ds_dp(use_none_as_null=True)

        dDeltaPsi_dp = []
        dpv_dp = []

        # loop through the parameters
        for der in dB_dxluc_p:

            if der is None:
                dpv_dp.append(None)
                dDeltaPsi_dp.append(None)
                continue

            der_mat = flex.mat3_double(len(U), der.elems)

            # calculate the derivative of q for this parameter
            dq = U * der_mat * h

            # calculate the derivative of r for this parameter
            dr = dq.rotate_around_origin(e1, DeltaPsi)

            # calculate the derivative of DeltaPsi for this parameter
            dDeltaPsi = -1.0 * (dr.dot(s1)) / (e1.cross(r).dot(s0))
            dDeltaPsi_dp.append(dDeltaPsi)

            # derivative of the axis e1
            q_dot_dq = q.dot(dq)
            dq0 = (q_scalar * dq - (q_dot_dq * q0)) / qq
            de1_dp = dq0.cross(s0u)

            # calculate (d[r]/d[e1])(d[e1]/dp)
            # dr_de1 = dRq_de(DeltaPsi, e1, q)
            # drde_dedp = dr_de1 * de1_dp

            # The above calculation is not correct. Instead, calculate the partial
            # derivative of r wrt change in rotation axis e1 by finite differences
            dp = 1.0e-8
            del_e1 = de1_dp * dp
            e1f = e1 + del_e1 * 0.5
            rfwd = q.rotate_around_origin(e1f, DeltaPsi)
            e1r = e1 - del_e1 * 0.5
            rrev = q.rotate_around_origin(e1r, DeltaPsi)
            drde_dedp = (rfwd - rrev) * (1 / dp)

            # calculate the derivative of pv for this parameter
            dpv = D * (dr + e1.cross(r) * dDeltaPsi + drde_dedp)
            dpv_dp.append(dpv)

        return dpv_dp, dDeltaPsi_dp

    @staticmethod
    def _calc_dX_dp_and_dY_dp_from_dpv_dp(w_inv, u_w_inv, v_w_inv, dpv_dp):
        """helper function to calculate positional derivatives from
        dpv_dp using the quotient rule"""

        dX_dp = []
        dY_dp = []

        for der in dpv_dp:
            if der is None:
                dX_dp.append(None)
                dY_dp.append(None)
            else:
                du_dp, dv_dp, dw_dp = der.parts()

                dX_dp.append(w_inv * (du_dp - dw_dp * u_w_inv))
                dY_dp.append(w_inv * (dv_dp - dw_dp * v_w_inv))

        return dX_dp, dY_dp


class StillsPredictionParameterisationSparse(
    SparseGradientVectorMixin, StillsPredictionParameterisation
):
    """A version of StillsPredictionParameterisation that uses a sparse matrix
    data structure for memory efficiency when there are a large number of
    Experiments"""

    pass


class SphericalRelpStillsPredictionParameterisation(StillsPredictionParameterisation):
    """Modified StillsPredictionParameterisation for the model that assumes
    relps are spherical and prediction requires that this sphere intersects the
    Ewald sphere, not that the relp centre is rotated onto the Ewald sphere.
    Gradients of DeltaPsi are still calculated for use as a restraint."""

    def _local_setup(self, reflections):
        """Setup additional attributes used in gradients calculation. These are
        specific to the stills-type prediction parameterisations that leaves the
        relp q generally off the Ewald sphere but assumes it has spherical extent
        which intersects the Ewald sphere. Gradients of the minimum rotation
        angle DeltaPsi are still calculated for use as a restraint"""

        self._DeltaPsi = reflections["delpsical.rad"]

        # q is the reciprocal lattice vector, in the lab frame
        self._q = self._UB * self._h

        # quantities involving the direct beam vector
        self._q_s0 = self._q + self._s0
        s = self._q_s0.norms()
        ss = self._q_s0.dot(self._q_s0)
        sss = s * ss
        self._inv_s = 1.0 / s
        self._inv_sss = 1.0 / sss

        # r is the reciprocal lattice vector rotated to the Ewald sphere, required
        # for derivatives of DeltaPsi
        self._r = self._s1 - self._s0

        # we also need the unit directions q0 and s0u
        self._q0 = self._q.each_normalize()
        self._s0u = self._s0.each_normalize()

        # e1 is the unit vector about which DeltaPsi rotation is defined
        self._e1 = self._q0.cross(self._s0u).each_normalize()

        # q1 completes an orthonormal set with q0 and e1
        self._q1 = self._q0.cross(self._e1).each_normalize()

        # c0 completes an orthonormal set with s0u and e1
        self._c0 = self._s0u.cross(self._e1).each_normalize()

        # we want the wavenumber and wavelength
        self._nu = self._s0.norms()
        self._wavelength = 1.0 / self._nu

        return

    def _beam_derivatives(self, isel, parameterisation=None, reflections=None):
        """helper function to extend the derivatives lists by derivatives of the
        beam parameterisations"""

        # Get required data
        s0 = self._s0.select(isel)
        s0u = self._s0u.select(isel)
        nu = self._nu.select(isel)
        r = self._r.select(isel)
        e1 = self._e1.select(isel)
        q_s0 = self._q_s0.select(isel)
        inv_s = self._inv_s.select(isel)
        inv_sss = self._inv_sss.select(isel)
        D = self._D.select(isel)

        # get the derivatives of the beam vector wrt the parameters
        ds0_dbeam_p = parameterisation.get_ds_dp(use_none_as_null=True)

        dDeltaPsi_dp = []
        dpv_dp = []

        # loop through the parameters
        for der in ds0_dbeam_p:

            if der is None:
                dpv_dp.append(None)
                dDeltaPsi_dp.append(None)
                continue

            # repeat the derivative in an array
            ds0 = flex.vec3_double(len(s0u), der.elems)

            # we need the derivative of the unit beam direction too. This requires
            # scaling by the wavelength and projection onto the Ewald sphere
            # scaled = ds0 * wl
            # ds0u = scaled.dot(c0) * c0 + scaled.dot(e1) * e1

            # calculate the derivative of DeltaPsi for this parameter
            dDeltaPsi = -1.0 * (r.dot(ds0)) / (e1.cross(r).dot(s0))
            dDeltaPsi_dp.append(dDeltaPsi)

            # term 1
            term1 = (s0u.dot(ds0) * q_s0 + nu * ds0) * inv_s

            # term 2
            term2 = (nu * q_s0 * q_s0.dot(ds0)) * inv_sss

            # calculate the derivative of pv for this parameter
            dpv_dp.append(D * (term1 - term2))

        return dpv_dp, dDeltaPsi_dp

    def _xl_orientation_derivatives(
        self, isel, parameterisation=None, reflections=None
    ):
        """helper function to extend the derivatives lists by
        derivatives of the crystal orientation parameterisations"""

        # Get required data
        B = self._B.select(isel)
        h = self._h.select(isel)
        e1 = self._e1.select(isel)
        DeltaPsi = self._DeltaPsi.select(isel)
        s1 = self._s1.select(isel)
        nu = self._nu.select(isel)
        q_s0 = self._q_s0.select(isel)
        inv_s = self._inv_s.select(isel)
        inv_sss = self._inv_sss.select(isel)
        r = self._r.select(isel)
        s0 = self._s0.select(isel)
        D = self._D.select(isel)

        # get derivatives of the U matrix wrt the parameters
        dU_dxlo_p = parameterisation.get_ds_dp(use_none_as_null=True)

        dDeltaPsi_dp = []
        dpv_dp = []

        # loop through the parameters
        for der in dU_dxlo_p:

            if der is None:
                dpv_dp.append(None)
                dDeltaPsi_dp.append(None)
                continue

            der_mat = flex.mat3_double(len(B), der.elems)

            # calculate the derivative of q for this parameter
            dq = der_mat * B * h

            # calculate the derivative of r for this parameter
            dr = dq.rotate_around_origin(e1, DeltaPsi)

            # calculate the derivative of DeltaPsi for this parameter
            dDeltaPsi = -1.0 * (dr.dot(s1)) / (e1.cross(r).dot(s0))
            dDeltaPsi_dp.append(dDeltaPsi)

            # term 1
            term1 = (nu * dq) * inv_s

            # term 2
            term2 = (nu * q_s0 * q_s0.dot(dq)) * inv_sss

            # calculate the derivative of pv for this parameter
            dpv_dp.append(D * (term1 - term2))

        return dpv_dp, dDeltaPsi_dp

    def _xl_unit_cell_derivatives(self, isel, parameterisation=None, reflections=None):
        """helper function to extend the derivatives lists by
        derivatives of the crystal unit cell parameterisations"""

        # Get required data
        U = self._U.select(isel)
        h = self._h.select(isel)
        e1 = self._e1.select(isel)
        DeltaPsi = self._DeltaPsi.select(isel)
        s1 = self._s1.select(isel)
        nu = self._nu.select(isel)
        q_s0 = self._q_s0.select(isel)
        inv_s = self._inv_s.select(isel)
        inv_sss = self._inv_sss.select(isel)
        r = self._r.select(isel)
        s0 = self._s0.select(isel)
        D = self._D.select(isel)

        # get derivatives of the B matrix wrt the parameters
        dB_dxluc_p = parameterisation.get_ds_dp(use_none_as_null=True)

        dDeltaPsi_dp = []
        dpv_dp = []

        # loop through the parameters
        for der in dB_dxluc_p:

            if der is None:
                dpv_dp.append(None)
                dDeltaPsi_dp.append(None)
                continue

            der_mat = flex.mat3_double(len(U), der.elems)

            # calculate the derivative of q for this parameter
            dq = U * der_mat * h

            # calculate the derivative of r for this parameter
            dr = dq.rotate_around_origin(e1, DeltaPsi)

            # calculate the derivative of DeltaPsi for this parameter
            dDeltaPsi = -1.0 * (dr.dot(s1)) / (e1.cross(r).dot(s0))
            dDeltaPsi_dp.append(dDeltaPsi)

            # term 1
            term1 = (nu * dq) * inv_s

            # term 2
            term2 = (nu * q_s0 * q_s0.dot(dq)) * inv_sss

            # calculate the derivative of pv for this parameter
            dpv = D * (term1 - term2)
            dpv_dp.append(dpv)

        return dpv_dp, dDeltaPsi_dp


class SphericalRelpStillsPredictionParameterisationSparse(
    SparseGradientVectorMixin, SphericalRelpStillsPredictionParameterisation
):
    """A version of SphericalRelpStillsPredictionParameterisation that uses a
    sparse matrix data structure for memory efficiency when there are a large
    number of Experiments"""

    pass



"""
Test command line program dials.two_theta_refine by running a job with saved
data and comparing with expected output.
"""


from __future__ import annotations

from pathlib import Path

import procrunner
import pytest

from dxtbx.model.experiment_list import ExperimentListFactory


def test(dials_data, tmpdir):
    """Test two theta refine on integrated data."""
    # use multiple scan small molecule data for this test
    data_dir = dials_data("l_cysteine_dials_output", pathlib=True)
    prefix = (20, 25, 30, 35)
    exp_path = [data_dir / ("%d_integrated_experiments.json" % p) for p in prefix]
    pkl_path = [data_dir / ("%d_integrated.pickle" % p) for p in prefix]

    for pth in exp_path + pkl_path:
        assert pth.is_file(), f"{str(pth)} missing"

    cmd = (
        [
            "dials.two_theta_refine",
            "cif=refined_cell.cif",
            "output.correlation_plot.filename=corrplot.png",
        ]
        + [str(p) for p in exp_path]
        + [str(p) for p in pkl_path]
    )

    print(cmd)

    # work in a temporary directory
    result = procrunner.run(cmd, working_directory=tmpdir)
    assert not result.returncode and not result.stderr
    assert Path(tmpdir / "refined_cell.expt").is_file()
    ref_exp = ExperimentListFactory.from_json_file(
        str(tmpdir / "refined_cell.expt"), check_format=False
    )

    xls = ref_exp.crystals()
    assert len(xls) == 4
    for xl in xls:
        assert xl.get_unit_cell() != xl.get_recalculated_unit_cell()
        # test refined crystal model against expected values
        assert xl.get_recalculated_unit_cell().parameters() == pytest.approx(
            (5.428022880, 8.144145476, 12.039666971, 90.0, 90.0, 90.0), 1e-4
        )
        assert xl.get_recalculated_cell_parameter_sd() == pytest.approx(
            (9.58081e-5, 0.000149909, 0.000215765, 0, 0, 0), 1e-4
        )
        assert xl.get_recalculated_cell_volume_sd() == pytest.approx(0.0116254298, 1e-4)


def test_two_theta_refine_scaled_data(dials_data, tmpdir):
    """Test two theta refine on scaled data."""
    location = dials_data("l_cysteine_4_sweeps_scaled", pathlib=True)
    refls = str(location / "scaled_20_25.refl")
    expts = str(location / "scaled_20_25.expt")

    command = [
        "dials.two_theta_refine",
        refls,
        expts,
        "output.experiments=refined_cell.expt",
        "partiality_threshold=0.99",
    ]
    result = procrunner.run(command, working_directory=tmpdir)
    assert not result.returncode and not result.stderr
    assert Path(tmpdir / "refined_cell.expt").is_file()

    ref_exp = ExperimentListFactory.from_json_file(
        str(tmpdir / "refined_cell.expt"), check_format=False
    )

    assert len(ref_exp.crystals()) == 2
    for xl in ref_exp.crystals():
        # test refined crystal model against expected values
        assert xl.get_recalculated_unit_cell().parameters() == pytest.approx(
            (5.426921, 8.146654, 12.037366, 90.0, 90.0, 90.0), 1e-4
        )
        assert xl.get_recalculated_cell_parameter_sd() == pytest.approx(
            (2.0123e-04, 2.8039e-04, 4.5284e-04, 0, 0, 0), 1e-4
        )
        assert xl.get_recalculated_cell_volume_sd() == pytest.approx(0.0237364, 1e-4)



from __future__ import annotations

import procrunner

from dxtbx.serialize import load


def test(dials_data, tmp_path):
    experiments = dials_data("centroid_test_data", pathlib=True) / "experiments.json"

    # Patched data file. Original had trusted_range from -1, but now this range
    # is defined to start from the minimum trusted value. This test should be
    # updated with new data.
    # https://github.com/dials/dials/issues/2200
    exp = load.experiment_list(experiments)
    panel = exp[0].detector[0]
    max_trusted = panel.get_trusted_range()[1]
    panel.set_trusted_range((0, max_trusted))
    exp.as_json(tmp_path / "trusted_range_patch.expt")

    result = procrunner.run(
        [
            "dials.background",
            "output.plot=background.png",
            "image=1",
            "trusted_range_patch.expt",
        ],
        working_directory=tmp_path,
    )

    assert not result.returncode and not result.stderr
    assert (tmp_path / "background.png").is_file()

    for line in result.stdout.splitlines():
        if line.startswith(b"Mean background"):
            assert line.endswith(b"0.559")
            break


def test_checkpoints(dials_data, tmp_path):
    experiments = dials_data("centroid_test_data", pathlib=True) / "experiments.json"

    # Patched data file. Original had trusted_range from -1, but now this range
    # is defined to start from the minimum trusted value. This test should be
    # updated with new data.
    # https://github.com/dials/dials/issues/2200
    exp = load.experiment_list(experiments)
    panel = exp[0].detector[0]
    max_trusted = panel.get_trusted_range()[1]
    panel.set_trusted_range((0, max_trusted))
    exp.as_json(tmp_path / "trusted_range_patch.expt")

    result = procrunner.run(
        [
            "dials.background",
            "n_checkpoints=3",
            "trusted_range_patch.expt",
        ],
        working_directory=tmp_path,
    )

    assert not result.returncode and not result.stderr
    images = [
        line.decode()
        for line in result.stdout.splitlines()
        if b"For imageset 0" in line
    ]
    images = [
        line.replace("For imageset 0 image", "").replace(":", "") for line in images
    ]
    images = {int(e) for e in images}
    assert len(images) == 3
    assert sorted(images) == [1, 5, 9]


def test_multiple_imagesets(dials_data, tmp_path):
    filenames = sorted(
        dials_data("thaumatin_grid_scan", pathlib=True).glob("thau_3_2_00*.cbf.bz2")
    )
    filenames.extend(
        sorted(dials_data("centroid_test_data", pathlib=True).glob("centroid_*.cbf"))
    )

    result = procrunner.run(
        [
            "dials.background",
            "output.plot=background.png",
            "image=1,2",
            "size_inches=16,10",
        ]
        + filenames,
        working_directory=tmp_path,
    )

    assert not result.returncode and not result.stderr
    assert (tmp_path / "background.png").is_file()

    lines = result.stdout.splitlines()
    assert b"For imageset 0 image 1:" in lines
    assert b"For imageset 0 image 2:" in lines
    assert b"For imageset 1 image 1:" in lines
    assert b"For imageset 1 image 2:" in lines



from __future__ import annotations

import pytest

from cctbx import crystal, miller
from scitbx.array_family import flex

from dials.algorithms.merging.reporting import dano_over_sigdano


def test_dano_over_sigdano():
    """Create anomalous difference data and check the calculated value."""
    ms = miller.build_set(
        crystal_symmetry=crystal.symmetry(
            space_group_symbol="P222", unit_cell=(6, 6, 6, 90, 90, 90)
        ),
        anomalous_flag=True,
        d_min=5.0,
    ).expand_to_p1()
    ma = miller.array(
        ms, data=flex.double([1, 2, 1, 3, 1, 4]), sigmas=flex.double(6, 1)
    )
    # differences are (1, 2, 3) i.e. mean 2, sigmas (sqrt2, sqrt2, sqrt2)
    assert dano_over_sigdano(ma) == pytest.approx(2**0.5)



"""
Tests for the functions in dials.util.options
"""

from __future__ import annotations

from unittest.mock import Mock

import pytest

from dxtbx.model import Experiment, ExperimentList

from dials.util import Sorry
from dials.util.options import (
    ArgumentParser,
    flatten_reflections,
    reflections_and_experiments_from_files,
)

from . import mock_reflection_file_object, mock_two_reflection_file_object


def test_cannot_read_headerless_h5(dials_data):
    data_h5 = dials_data("vmxi_thaumatin", pathlib=True) / "image_15799_data_000001.h5"
    parser = ArgumentParser(read_experiments_from_images=True)
    with pytest.raises(Sorry):
        parser.parse_args([str(data_h5)])


def test_flatten_experiments_updating_id_values():
    """Test the correct handling of duplicate table id values.

    Note that this function does not have the ability to update the
    experiment string identifier, only ensure that the table id values
    do not clash (it is not possible even to load multiple experiments
    with the same identifier).
    """
    # Test the case of two single reflection tables.
    file_list = [mock_reflection_file_object(id_=0), mock_reflection_file_object(id_=0)]
    rs = flatten_reflections(file_list)
    assert rs[0] is file_list[0].data
    assert list(rs[0]["id"]) == [-1, 0, 0]
    assert list(rs[0].experiment_identifiers().keys()) == [0]
    assert list(rs[0].experiment_identifiers().values()) == ["0"]
    assert rs[1] is file_list[1].data
    assert list(rs[1]["id"]) == [-1, 1, 1]
    assert list(rs[1].experiment_identifiers().keys()) == [1]
    assert list(rs[1].experiment_identifiers().values()) == ["0"]

    # Now test the case where one reflection table contains two experiments
    file_list = [mock_two_reflection_file_object(), mock_reflection_file_object(id_=0)]
    rs = flatten_reflections(file_list)
    assert rs[0] is file_list[0].data
    assert list(rs[0]["id"]) == [-1, 0, 0, 1, 1]
    assert list(rs[0].experiment_identifiers().keys()) == [0, 1]
    assert list(rs[0].experiment_identifiers().values()) == ["0", "2"]
    assert rs[1] is file_list[1].data
    assert list(rs[1]["id"]) == [-1, 2, 2]
    assert list(rs[1].experiment_identifiers().keys()) == [2]
    assert list(rs[1].experiment_identifiers().values()) == ["0"]

    file_list = [
        mock_reflection_file_object(id_=0),
        mock_two_reflection_file_object(ids=[1, 2]),
    ]
    rs = flatten_reflections(file_list)
    assert rs[0] is file_list[0].data
    assert list(rs[0]["id"]) == [-1, 0, 0]
    assert list(rs[0].experiment_identifiers().keys()) == [0]
    assert list(rs[0].experiment_identifiers().values()) == ["0"]
    assert rs[1] is file_list[1].data
    assert list(rs[1]["id"]) == [-1, 1, 1, 2, 2]
    assert list(rs[1].experiment_identifiers().keys()) == [1, 2]
    assert list(rs[1].experiment_identifiers().values()) == ["1", "2"]


def test_reflections_and_experiments_from_files():
    """Test correct extracting of reflections and experiments."""
    # Test when input reflections order matches the experiments order
    refl_file_list = [
        mock_two_reflection_file_object(ids=[0, 1]),
        mock_reflection_file_object(id_=2),
    ]

    def mock_exp_obj(id_=0):
        """Make a mock experiments file object."""
        exp = Mock()
        exp.data = ExperimentList()
        exp.data.append(Experiment(identifier=str(id_)))
        return exp

    exp_file_list = [mock_exp_obj(id_=i) for i in [0, 1, 2]]

    refls, expts = reflections_and_experiments_from_files(refl_file_list, exp_file_list)
    assert refls[0] is refl_file_list[0].data
    assert refls[1] is refl_file_list[1].data
    assert expts[0].identifier == "0"
    assert expts[1].identifier == "1"
    assert expts[2].identifier == "2"

    # Test when input reflections order does not match experiments order.
    refl_file_list = [
        mock_reflection_file_object(id_=2),
        mock_two_reflection_file_object(ids=[0, 1]),
    ]
    refls, expts = reflections_and_experiments_from_files(refl_file_list, exp_file_list)
    assert refls[0] is refl_file_list[1].data
    assert refls[1] is refl_file_list[0].data
    assert expts[0].identifier == "0"
    assert expts[1].identifier == "1"
    assert expts[2].identifier == "2"
